""" Constants from constants.h from ADMS, or CEDIC simul to be more exact!

Namings :

*  M is a mathematical constant
*  P is a physical constant. The constants have been taken from http://physics.nist.gov
*  others: HICUM constants

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


M_E = 2.7182818284590452354
""" define M_E 2.7182818284590452354 """

M_LOG2E = 1.4426950408889634074
""" define M_LOG2E 1.4426950408889634074 """

M_LOG10E = 0.43429448190325182765
""" define M_LOG10E 0.43429448190325182765 """

M_LN2 = 0.69314718055994530942
""" define M_LN2 0.69314718055994530942 """

M_LN10 = 2.30258509299404568402
""" define M_LN10 2.30258509299404568402 """

M_PI = 3.14159265358979323846
""" define M_PI 3.14159265358979323846 """

M_TWO_PI = 6.28318530717958647652
""" define M_TWO_PI 6.28318530717958647652 """

M_PI_2 = 1.57079632679489661923
""" define M_PI_2 1.57079632679489661923 """

M_PI_4 = 0.78539816339744830962
""" define M_PI_4 0.78539816339744830962 """

M_1_PI = 0.31830988618379067154
""" define M_1_PI 0.31830988618379067154 """

M_2_PI = 0.63661977236758134308
""" define M_2_PI 0.63661977236758134308 """

M_2_SQRTPI = 1.12837916709551257390
""" define M_2_SQRTPI 1.12837916709551257390 """

M_SQRT2 = 1.41421356237309504880
""" define M_SQRT2 1.41421356237309504880 """

M_SQRT1_2 = 0.70710678118654752440
""" define M_SQRT1_2 0.70710678118654752440 """


P_Q = 1.602176462e-19
"""  charge of electron in coulombs 1.602176462e-19 """

P_C = 2.99792458e8
"""  speed of light in vacuum in meters/sec 2.99792458e8 """

P_K = 1.3806503e-23
"""  Boltzmann's constant in joules/kelvin 1.3806503e-23 """

P_H = 6.62606876e-34
"""  Planck's constant in joules*sec 6.62606876e-34 """

P_EPS0 = 8.854187817e-12
""""  permittivity of vacuum in farads/meter 8.854187817e-12 """

P_U0 = 12.566370614e-7
""""  permeability of vacuum in henrys/meter (4.0e-7 * M_PI) (12.566370614e-7) """

P_CELSIUS0 = 273.15
"""  zero celsius in kelvin 273.15 """

VPT_thresh = 1.0e2
Dexp_lim = 80.0
Cexp_lim = 80.0
DFa_fj = 1.921812
DFal_qr = 0.01
RTOLC = 1.0e-5
l_itmax = 100
TMAX = 326.85
TMIN = -100.0
LN_EXP_LIMIT = 11.0
MIN_R = 0.001
Gmin = 1.0e-12


def calc_VT(temp):
    """Calculates the thermal voltage for the given temperature (in Kelvin)"""
    return P_K * temp / P_Q


vT_300 = calc_VT(300)
""" thermal voltage for 300 Kelvin """

DA_FI = 1.921812
""" constant smoothing parameter for HICUM """



""" test ADS input file generation, copy to server, run simulation and copy back

USING VerilogAE to obtain the SPG modelcard!
-> needs to be installed and call:

verilogae install --module-name sgp_v1p0 DMT_other/DMT/hl2/sgp_v1p0_vae.va

"""
import types
import logging

from pathlib import Path
from DMT.core import DutType, Sweep, specifiers, SimCon, Plot, MCard
from DMT.core.circuit import SGP_BJT
from DMT.hl2 import VA_FILES, sgp_default_circuits
from DMT.xyce import DutXyce
from DMT.spectre import DutSpectre

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_xyce_vs_spectre.log",
    filemode="w",
)


def get_circuit(self):
    return sgp_default_circuits.get_circuit(self.default_circuit, self)


def get_spg_mc():
    modelcard = MCard(
        sgp_v1p0.nodes,
        "QSGP",
        SGP_BJT,
        1.1,
        vae_module=sgp_v1p0,
        default_circuit="common_emitter",
    )
    modelcard.default_module_name = SGP_BJT  # it is set from the vae_module to something different :/ -> this is intentional and usually pretty good
    modelcard.get_circuit = types.MethodType(
        get_circuit, modelcard
    )  # we need to bind here, this is a little tricky -.-
    modelcard.load_model_parameters(folder_path / "bjt.lib")

    # modelcard.set_values(
    #     {"is": 1e-16,}
    # )

    return modelcard


def get_dut_xyce_build_in():
    modelcard = get_spg_mc()

    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_BI_",
        nodes="C,B,E",
        copy_va_files=False,
        reference_node="E",
    )


def get_dut_xyce_va():
    modelcard = get_spg_mc()
    modelcard.va_file = VA_FILES["SGPV1.0"]
    modelcard.vae_module = sgp_v1p0  # parse!
    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_VA_",
        nodes="C,B,E",
        copy_va_files=True,
        reference_node="E",
    )


def get_dut_spectre_build_in():
    modelcard = get_spg_mc()
    return DutSpectre(
        None,
        DutType.npn,
        modelcard,
        name="Spec_BI_",
        nodes="C,B,E",
        copy_va_files=False,
        reference_node="E",
        simulate_on_server=True,
    )


def get_dut_spectre_va():
    modelcard = get_spg_mc()
    modelcard.va_file = VA_FILES["SGPV1.0"]
    modelcard.vae_module = sgp_v1p0  # parse!
    return DutSpectre(
        None,
        DutType.npn,
        modelcard,
        name="Spec_VA_",
        nodes="C,B,E",
        copy_va_files=True,
        reference_node="E",
        simulate_on_server=True,
    )


def get_sweep():
    sweepdef = [
        {"var_name": "FREQ", "sweep_order": 4, "sweep_type": "LOG", "value_def": [8, 9, 2]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0, 1, 21],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 3,
            "sweep_type": "SYNC",
            "master": specifiers.VOLTAGE + "B",
            "offset": specifiers.VOLTAGE + ["C", "B"],
        },
        {
            "var_name": specifiers.VOLTAGE + ["C", "B"],
            "sweep_order": 2,
            "sweep_type": "LIST",
            "value_def": [0.3, 0, -0.3],
            # "value_def": [0],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    return Sweep(
        "gummel",
        sweepdef=sweepdef,
        outputdef=[specifiers.CURRENT + "C", specifiers.CURRENT + "B"],
        othervar={"TEMP": 300},
    )


def run_and_read(dut_test, sweep_test):
    sim_con = SimCon(t_max=80)  # damn VA-Compile time
    sim_con.append_simulation(dut=dut_test, sweep=sweep_test)
    sim_con.run_and_read(force=True)


if __name__ == "__main__":
    duts = []
    duts.append(get_dut_spectre_build_in())
    # duts.append(get_dut_xyce_build_in())
    # duts.append(get_dut_xyce_va())
    duts.append(get_dut_spectre_va())
    sweep = get_sweep()

    run_and_read(duts, sweep)

    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    plt_gummel = Plot(
        r"$I_{\mathrm{C}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{C}}$",
        y_log=True,
        style="xtraction_color",
    )
    plt_ib = Plot(
        r"$I_{\mathrm{B}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{B}}$",
        y_log=True,
        style="xtraction_color",
    )

    for dut in duts:
        dut_name = dut.name[:7]
        df = dut.get_data(sweep=sweep)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)

        for _index, vbc, data in df.iter_unique_col(col_vbc, decimals=3):
            data = data[data[specifiers.FREQUENCY] == 1e8]

            plt_gummel.add_data_set(
                data[col_vbe],
                data[col_ic],
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )
            plt_ib.add_data_set(
                data[col_vbe],
                data[col_ib],
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )

    plt_ib.plot_pyqtgraph(show=False)
    plt_gummel.plot_pyqtgraph(show=True)



import pytest
from pathlib import Path
from DMT.core import DataFrame
from DMT.core import (
    specifiers,
    sub_specifiers,
    get_specifier_from_string,
    set_col_name,
    SpecifierStr,
)
from DMT.core import DatabaseManager

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent


def test_index_objects():
    new_spec = SpecifierStr("NEWSPECVAL")
    new_sub_spec = SpecifierStr("", sub_specifiers="NEWSUBSPECVAL")

    specifiers.add_members([("NEWSPECNAME", new_spec)])
    sub_specifiers.add_members([("NEWSUBSPECNAME", new_sub_spec)])

    # access them again:
    new_spec_2 = specifiers.NEWSPECNAME
    assert new_spec == new_spec_2

    # work with them:
    assert new_spec_2 + sub_specifiers.NEWSUBSPECNAME == "NEWSPECVAL|NEWSUBSPECVAL"

    with pytest.raises(OSError):
        specifiers.add_members([("CURRENT", SpecifierStr("I"))])
    with pytest.raises(OSError):
        specifiers.add_members([("CURRENT", SpecifierStr("B"))])
    with pytest.raises(OSError):
        sub_specifiers.add_members([("AREA", SpecifierStr("", sub_specifiers="AREA2"))])


def test_specifier_texts():
    assert specifiers.VOLTAGE + "B" == "V_B"
    assert specifiers.VOLTAGE + "_B" == "V_B"
    vb = specifiers.VOLTAGE + "B"
    vbe = vb + "E"
    assert vbe == "V_BE"
    assert vbe.to_raw_string() == "V_BE"
    assert isinstance(vbe.to_raw_string(), str)

    assert "V" + SpecifierStr("", "B") == "V_B"
    assert vbe + "" == vbe
    assert vbe is not None  # valid comparision !
    assert vbe != None  # valid comparision !

    with pytest.raises(TypeError):
        _a = vbe + 5
    with pytest.raises(TypeError):
        _a = 5 + vbe

    assert specifiers.VOLTAGE + "B" + sub_specifiers.AREA == "V_B|AREA"
    assert (
        specifiers.VOLTAGE + "B" + sub_specifiers.AREA + sub_specifiers.FORCED == "V_B|AREA|FORCED"
    )
    assert (
        specifiers.VOLTAGE + "B" + [sub_specifiers.AREA, sub_specifiers.FORCED] == "V_B|AREA|FORCED"
    )
    assert (
        specifiers.VOLTAGE + "B" + [sub_specifiers.FORCED, sub_specifiers.AREA] == "V_B|AREA|FORCED"
    )  # order of sub_specifiers does not matter
    assert (
        specifiers.VOLTAGE + "B" + {sub_specifiers.AREA, sub_specifiers.FORCED} == "V_B|AREA|FORCED"
    )  # can also be a set
    assert specifiers.VOLTAGE + "B" + sub_specifiers.FORCED == "V_B|FORCED"
    assert specifiers.VOLTAGE + sub_specifiers.FORCED == "V|FORCED"

    assert set_col_name(specifiers.SS_PARA_Y, "2", "1") == "Y_21"
    assert specifiers.SS_PARA_Y + ["2", "1"] == "Y_21"

    # order of sub_specifiers does not matter
    assert specifiers.VOLTAGE + "B" + [
        sub_specifiers.AREA,
        sub_specifiers.FORCED,
    ] == specifiers.VOLTAGE + "B" + [sub_specifiers.FORCED, sub_specifiers.AREA]


def test_specifier_from_string():
    assert get_specifier_from_string("TAU_F").specifier == "TAU_F"
    assert get_specifier_from_string("BETA").specifier == "BETA"
    assert get_specifier_from_string("V_BE", nodes=["B", "E"]).specifier == "V"
    assert get_specifier_from_string("V_BE", nodes=["B", "E"]).nodes[1] == "E"
    assert sub_specifiers.FORCED in get_specifier_from_string("V_BE|FORCED", nodes=["B", "E"])
    assert (
        sub_specifiers.FORCED.sub_specifiers
        <= get_specifier_from_string("V_BE|FORCED", nodes=["B", "E"]).sub_specifiers
    )
    # frozenset comparisions
    # https://www.geeksforgeeks.org/sets-in-python/

    assert specifiers.FREQUENCY == "FREQ"
    assert set_col_name(specifiers.FREQUENCY) == "FREQ"
    assert SpecifierStr.string_from_load(specifiers.FREQUENCY.string_to_save()) == "FREQ"

    # only works for marked strings
    assert isinstance(SpecifierStr.string_from_load("FREQ"), str)
    assert not isinstance(SpecifierStr.string_from_load("FREQ"), SpecifierStr)


def column_save_load():
    a = DataFrame({SpecifierStr("V", "B"): [1]})

    dbm = DatabaseManager()
    dbm.save_df(a, test_path / "tmp" / "test_df.p")
    b = dbm.load_df(test_path / "tmp" / "test_df.p")

    assert isinstance(a.columns[0], SpecifierStr)
    assert isinstance(b.columns[0], SpecifierStr)

    data = {"a": a, "b": b}
    dbm.save_db(test_path / "tmp" / "test_db.p", data)
    data_loaded = dbm.load_db(test_path / "tmp" / "test_db.p")

    assert isinstance(data_loaded["a"].columns[0], SpecifierStr)
    assert data_loaded["a"].columns[0] == data["a"].columns[0]


def test_pretty_printing():
    col_ic = specifiers.CURRENT + "C"
    # description for POA
    assert col_ic.get_descriptor() == "current"

    # unit:
    assert str(col_ic.get_pint_unit()) == "ampere"

    assert col_ic.get_tex_unit() == "\\si{\\ampere}"

    with pytest.raises(IOError):
        specifiers.ELECTRONS.get_pint_unit()

    # labeling
    assert col_ic.to_tex() == "I_{\\mathrm{C}}"
    assert col_ic.to_label() == "$I_{\\mathrm{C}}\\left(\\si{\\ampere}\\right)$"
    assert col_ic.to_label(negative=True) == "$-I_{\\mathrm{C}}\\left(\\si{\\ampere}\\right)$"

    col_y21 = specifiers.SS_PARA_Y + ["C", "B"]
    assert col_y21.to_tex() == "\\underline{Y}_{\\mathrm{21}}"
    col_y21_imag = specifiers.SS_PARA_Y + ["C", "B"] + sub_specifiers.IMAG
    assert col_y21_imag.to_tex() == "\\Im\\{\\underline{Y}_{\\mathrm{21}}\\}"


if __name__ == "__main__":
    test_index_objects()
    test_specifier_texts()
    test_specifier_from_string()
    column_save_load()
    test_pretty_printing()
    dummy = 1



""" Manges a device under test which can be simulated by Hdev using the TCAD Interface class.

Author: Markus Müller
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

# pylint: disable=no-member
import re
import copy
import os
import numpy as np
import pandas as pd
import logging
import h5py
from typing import cast
import shutil

from DMT.core import (
    DutTcad,
    DataFrame,
    specifiers,
    sub_specifiers,
    Sweep,
    read_data,
    get_specifier_from_string,
    constants,
    DutType,
)
from DMT.config import DATA_CONFIG

from DMT.exceptions import SimulationUnsuccessful
from DMT.config import COMMANDS
from hdevpy.tools import makeinp, turn_off_recombination

try:
    from HdevInterface import hdev_py
except ImportError:
    print("Warning hdevpy not available.")

# Hdev iv output columns not needed in current DMT framework
hdev_iv_fallback = {
    "T_LATTICE": None,
    "T_CPU": None,
}


class DutHdev(DutTcad):
    r"""Manages simulations and simulation data with Hdev.

    Parameters
    ----------
    database_dir : str
        Directory for the current databases
    dut_type   : :class:`~DMT.core.dut_type.DutType`
    inp_structure : dictionary
        Dictionary that describes the simulation domain according to the Hdev manual.
    name : str, {'dev'}, optional
        Prefix for the files
    nodes : str, optional
        Comma separated node list
    reference_node : str, optional
        Contact node, required reference for all output voltages. HAS TO BE SET BEFORE SIMULATION!
    simulator_command : str, optional
        System command to call the simulation. If none, config.py command is used.
    simulator_arguments : list[str], optional
        Arguments for the system command

    Attributes
    ----------
    sim_name : str
        DEVICE simulation title. Sort of a DUT-Name.

    """
    inited = False

    def __init__(
        self, database_dir, dut_type, inp_structure, name="hdev_", simulator_command=None, **kwargs
    ):
        if simulator_command is None:
            simulator_command = COMMANDS["Hdev"]

        super().__init__(
            database_dir,
            name,
            dut_type,
            inp_structure,
            simulator_command=simulator_command,
            inp_name="hdev_inp.din",
            **kwargs,
        )

    def create_inp_header(self, inp_):
        """Creates the inp_header from the given parameters.

        All parameters are taken from obj_inp_device.

        Parameters
        ----------
        inp_ : dict of dicts that describes a Hdev input structure

        Returns
        -------
        inp_header : str
        """
        if not "BIAS_DEF" in inp_:
            inp_["BIAS_DEF"] = {}

        if not "OUTPUT" in inp_:
            inp_["OUTPUT"] = {}

        # to be consistent with the resolving of DMT sweepdef for Hdev simulation
        inp_["BIAS_DEF"]["list"] = 1
        inp_["OUTPUT"]["path"] = "output"
        inp_["OUTPUT"]["name"] = "dut"
        self.inp_dict = inp_
        inp = makeinp(inp_, file=None)
        return inp

    def make_input(self, sweep):
        sweep = copy.deepcopy(sweep)
        """Adds bias blocks to a Hdev heade file (string)"""
        inp_str = self._inp_header
        inp_str = cast(str, inp_str)
        # find frequency sweep def
        has_f_def = False
        for i, sub_sweep in enumerate(sweep.sweepdef):
            if sub_sweep.var_name == specifiers.FREQUENCY:
                has_f_def = True
                break
        if has_f_def:
            f_def = sweep.sweepdef.pop(i)  # type: ignore
        sweep.set_values()
        df = sweep.create_df()
        # add T sweep if not specified
        has_t_def = False
        for sub_sweep in sweep.sweepdef:
            if sub_sweep.var_name == specifiers.TEMPERATURE:
                has_t_def = True

        try:
            if not has_t_def:
                temp_bias_info = {
                    "cont_name": "'T'",
                    "bias_fun": "'LIN'",
                    "bias_val": "{0:f} {0:f} {1:f}".format(sweep.othervar["TEMP"], len(sweep.df)),
                }
        except KeyError:
            if not has_t_def:
                raise IOError("DutHdev: The temperature needs to be specified explicitly!")
        # copy sweep definition:

        path_bias = os.path.join(DATA_CONFIG["directories"]["simulation"], "bias.h5")

        f = h5py.File(path_bias, "w")
        for key, val in df.items():
            try:
                cont_name = key.nodes[0]
            except IndexError:
                cont_name = str(key)[0]
            except AttributeError:
                cont_name = str(key)[0]
            f.create_dataset(cont_name, data=val)
        f.close()
        self.list_copy.append(path_bias)
        # end save sweep definition
        for sub_sweep in sweep.sweepdef:
            try:
                cont_name = "'" + sub_sweep.var_name.nodes[0] + "'"
            except IndexError:
                cont_name = "'" + str(sub_sweep.var_name)[0] + "'"

            bias_info = {
                "bias_fun": "'HDF'",
                "cont_name": cont_name,
            }

            # add sweep definition to inp file
            bias_str = (
                "&BIAS_INFO "
                + "".join([name + "=" + str(value) + " " for name, value in bias_info.items()])
                + "/"
            )
            inp_str = inp_str + "\n" + bias_str

        # add frequency def
        if has_f_def:
            sub_sweep = f_def
            if sub_sweep.sweep_type == "CON":
                bias_fun = "'TAB'"
                ac_info = {
                    "port1": "'" + self.ac_ports[0] + "'",
                    "port2": "'" + self.ac_ports[1] + "'",
                    "sweep_type": bias_fun,
                    "freq_val": sub_sweep.value_def[0],
                }
                # add sweep definition to inp file
                bias_str = (
                    "&AC_INFO "
                    + "".join([name + "=" + str(value) + " " for name, value in ac_info.items()])
                    + "/"
                )
                inp_str = inp_str + "\n" + bias_str
            elif sub_sweep.sweep_type == "LIN":
                bias_fun = "'LIN'"
                ac_info = {
                    "port1": "'" + self.ac_ports[0] + "'",
                    "port2": "'" + self.ac_ports[1] + "'",
                    "sweep_type": bias_fun,
                    "freq_val": " ".join(["{0:3.2e}".format(val) for val in sub_sweep.value_def]),
                }
                # add sweep definition to inp file
                bias_str = (
                    "&AC_INFO "
                    + "".join([name + "=" + str(value) + " " for name, value in ac_info.items()])
                    + "/"
                )
                inp_str = inp_str + "\n" + bias_str
            else:
                raise NotImplementedError

        if not has_t_def:
            bias_str = (
                "&BIAS_INFO "
                + "".join([name + "=" + str(value) + " " for name, value in temp_bias_info.items()])
                + "/"
            )
            inp_str = inp_str + "\n" + bias_str

        inp_str = inp_str + "\n"
        return inp_str

    def validate_simulation_successful(self, sweep):
        """Checks if the simulation of the given sweep was successful.

        Parameters
        ----------
        sweep  : :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated.

        Raises
        ------
        NotImplementedError
            If the Dut is not a simulatable dut.
        SimulationUnsuccessful
            If the simulation output is not valid.
        FileNotFoundError
            If the sim log file does not exist.
        """
        # get sweep folder
        sim_folder = self.get_sim_folder(sweep)
        log_str = ""
        with open(os.path.join(sim_folder, "sim.log")) as log_file:
            try:
                log_str = log_file.read()
            except:
                return True

        # if not 'simulation finished' in log_str:
        if False:
            pattern = r"(error.*)"
            r1 = re.findall(pattern, log_str)
            print("Hdev error message is:")
            print(r1)
            print("\n")
            raise SimulationUnsuccessful(
                "Hdev Simulation of the sweep "
                + sweep.name
                + " with the hash "
                + sweep.get_hash()
                + " failed! An error was found! Director is "
                + sim_folder
                + "."
            )
        else:
            try:
                pattern = r"system time:(.*)"
                sim_time = re.findall(pattern, log_str)
                return sim_time[0]
            except (KeyError, IndexError) as e:
                # error during simulation, sim_time not printed
                raise SimulationUnsuccessful(
                    "Hdev Simulation of the sweep "
                    + sweep.name
                    + " with the hash "
                    + sweep.get_hash()
                    + " failed! An error was found! Director is "
                    + str(sim_folder)
                    + "."
                ) from e

    def get_df(self, hdf_data, group, key):
        """Return the pandas prepresentation of a Hdev hdf5 table."""
        if group is None:
            data_ = hdf_data[key]
        else:
            try:
                data_ = hdf_data[group + "/" + key]
            except KeyError:
                raise IOError

        cols = {}
        if group is None:
            for col, name in data_.attrs.items():
                cols[int(col.replace("col", ""))] = name.decode("ascii")
        else:
            group = hdf_data[group]
            for col, name in group.attrs.items():
                cols[int(col.replace("col", ""))] = name.decode("ascii")

        rows = np.linspace(0, data_.shape[0] - 1, data_.shape[0])
        try:
            columns = [cols[i] for i in range(len(cols))]
        except KeyError:
            dummy = 1
        return DataFrame(data_[:, : len(columns)], index=rows, columns=columns)

    def import_output_data(self, sweep, delete_sim_results=False):
        r"""Read the output files that have been produced while simulating sweep and attach them to self.db.

        This is done by scanning the input file for "\*.elpa" files and (for internal data) for "int" files.

        Parameters
        ----------
        sweep  :  :class:`DMT.core.sweep.Sweep`
            Sweep that has been simulated for the desired output files.

        Returns
        -------
        output_files : [string]
            List of strings that contain the full path to the output files.
        delete_sim_results : {False, True}, optional
            If True, the simulation folder is deleted after reading.

        Notes
        -----
        .. todo:: Add more column renaming and resorting here!
        """
        # list_elpa_files = re.findall(
        #     r"\&ELEC_DATA file_name='(.*?\.elpa)'",
        #     str_inp_file
        # )

        sim_folder = self.get_sim_folder(sweep)
        key = self.get_sweep_key(sweep)

        try:
            df_ = h5py.File(os.path.join(sim_folder, "simulation_data.h5"))  # hdf5 simulation data
            df_iv = self.get_df(df_, None, "iv")
            try:
                df_cap = self.get_df(df_, None, "cap")
            except:
                df_cap = None
            dfs_laplace = []
            if "laplace" in df_.keys():  # have laplace files
                laplace_keys = list(df_["laplace"].keys())
                for i in range(len(laplace_keys)):
                    dfs_laplace.append(self.get_df(df_, "laplace", laplace_keys[i]))
            dfs_inqu = []
            if "inqu" in df_.keys():  # have inqu files
                inqu_keys = list(df_["inqu"].keys())
                for i in range(len(inqu_keys)):
                    try:
                        dfs_inqu.append(self.get_df(df_, "inqu", "op" + str(i + 1)))
                    except:
                        pass
            dfs_acinqu = []
            if "acinqu" in df_.keys():  # have acinqu files
                acinqu_keys = list(df_["acinqu"].keys())
                for acinqu_key in acinqu_keys:
                    try:
                        dfs_acinqu.append(self.get_df(df_, "acinqu", acinqu_key))
                    except:
                        pass

            dfs_ac = []
            if "ac" in df_.keys():  # have ac files
                ac_keys = list(df_["ac"].keys())
                for i in range(len(ac_keys)):
                    dfs_ac.append(self.get_df(df_, "ac", "op" + str(i + 1)))

            df_.close()

            # add the inqu and acinqu files
            for i, df_inqu in enumerate(dfs_inqu):
                key_ = self.join_key(key, "op" + str(i + 1)) + "_inqu"
                self.data[key_] = df_inqu

            # add the inqu and acinqu files
            for i, df_laplace in enumerate(dfs_laplace):
                key_ = self.join_key(key, "laplace" + "_" + laplace_keys[i].upper())
                self.data[key_] = df_laplace

            # add the acinqu files
            for acinqu_key, df_acinqu in zip(acinqu_keys, dfs_acinqu):
                key_ = self.join_key(key, "acinqu_" + acinqu_key)
                self.data[key_] = df_acinqu

            # add the cap files
            if df_cap is not None:
                key_ = self.join_key(key, "cap")
                self.data[key_] = df_cap

            ac = False
            # read in the iv data
            pd.options.mode.chained_assignment = None  # default='warn'
            if len(dfs_ac) > 1:
                # first we sort by n_op
                dfs_temp = []
                n = 0
                for df_ac in dfs_ac:
                    # read in ac df
                    df_dc = df_iv.iloc[[n]]
                    # extend ac df with dc data
                    for col_dc in df_iv.columns:
                        if col_dc not in df_ac.columns:
                            df_ac.loc[:, col_dc] = df_iv.loc[n, col_dc]  # take over values from DC

                    # extend dc df with ac data
                    for col_ac in df_ac.columns:
                        if col_ac not in df_dc.columns:
                            df_dc.loc[:, col_ac] = 0  # set AC values to zero in DC

                    dfs_temp.append(df_dc)
                    dfs_temp.append(df_ac)

                    n = n + 1

                # create new big dataframe with DC and AC data
                df_iv = pd.concat(dfs_temp)

            # convert columns to specifiers
            df_iv = df_iv.real2cmplx()
            for _col in df_iv.columns:
                df_iv.rename(
                    columns={_col: get_specifier_from_string(_col, nodes=self.nodes)}, inplace=True
                )

            if not df_iv.columns.is_unique:
                df_iv = df_iv.loc[:, ~df_iv.columns.duplicated()]

            try:
                freqs = np.unique(df_iv[specifiers.FREQUENCY].to_numpy())
                ac = True
                if len(freqs) == 2:
                    df_iv = df_iv[df_iv[specifiers.FREQUENCY] == np.max(freqs)]

                # ensure some columns
                df_iv.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "E", ports=["B", "C"])
                df_iv.ensure_specifier_column(specifiers.CAPACITANCE + "B" + "C", ports=["B", "C"])
                df_iv.ensure_specifier_column(specifiers.TRANSIT_FREQUENCY, ports=["B", "C"])
                df_iv.ensure_specifier_column(specifiers.TRANSCONDUCTANCE, ports=["B", "C"])
                df_iv.ensure_specifier_column(
                    specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL, ports=["B", "C"]
                )
            except KeyError:
                pass

            if (
                ac and len(dfs_inqu) > 0
            ):  # only one frequency simulated => post process (maybe also check for 1D)
                if not len(freqs) == 2:
                    raise IOError
                # post processing
                # 1 general stuff
                x = dfs_inqu[0]["X"].to_numpy()
                d = dfs_inqu[0][specifiers.NET_DOPING].to_numpy()
                junctions = (
                    np.where(np.sign(d[:-1]) != np.sign(d[1:]))[0] + 1
                )  # detect sign changes
                if junctions != []:  # no junction found
                    xje = x[junctions[0]]
                    xjc = x[junctions[1]]

                    # sum of charge
                    qp = np.zeros(len(df_iv))
                    qn = np.zeros(len(df_iv))
                    for i_row, _row in enumerate(df_iv.iterrows()):
                        try:
                            key_inqu = next(
                                _key
                                for _key in self.data.keys()
                                if "_inqu" in _key and "op" + str(i_row + 1) in _key
                            )
                        except:
                            continue
                        df_inqu = self.data[key_inqu]
                        qn[i_row] = np.trapz(df_inqu["N"], df_inqu["X"])
                        qp[i_row] = np.trapz(df_inqu["P"], df_inqu["X"])

                    df_iv["Q|N"] = qn
                    df_iv["Q|P"] = qp

                    # where inqu df is required for every op
                    gm = np.array(df_iv[specifiers.TRANSCONDUCTANCE])
                    tau_e = np.ones(len(df_iv))
                    tau_be = np.ones(len(df_iv))
                    tau_b = np.ones(len(df_iv))
                    tau_c = np.ones(len(df_iv))
                    tau_bc = np.ones(len(df_iv))

                    try:

                        for i_row, row in enumerate(df_iv.iterrows()):
                            key_inqu = next(
                                _key
                                for _key in self.data.keys()
                                if "_inqu" in _key and "op" + str(i_row + 1) in _key
                            )
                            key_ac_inqu = next(
                                _key
                                for _key in self.data.keys()
                                if "acinqu" in _key
                                and "op" + str(i_row + 1) in _key
                                and "port1" in _key
                            )

                            # get the necessary data
                            df_ac_inqu_i = self.data[key_ac_inqu]
                            dn_dic = df_ac_inqu_i["re_d_n_x"].to_numpy() / gm[i_row + 1]
                            dp_dic = df_ac_inqu_i["re_d_p_x"].to_numpy() / gm[i_row + 1]
                            droh_dic = dp_dic - dn_dic
                            # transit time
                            changes = (
                                np.where(np.sign(droh_dic[:-1]) != np.sign(droh_dic[1:]))[0] + 1
                            )
                            index_be = changes[np.argmin(np.abs(changes - junctions[0]))]
                            index_bc = changes[np.argmin(np.abs(changes - junctions[1]))]
                            xbe = x[index_be]
                            xbc = x[index_bc]

                            tau = np.zeros(len(x))
                            for j, x_ in enumerate(x):
                                if x_ <= xbe:
                                    tau[j] += np.trapz(dp_dic[:j], x[:j])
                                    tau[j] += np.trapz(dn_dic[:j], x[:j]) - np.trapz(
                                        dp_dic[:j], x[:j]
                                    )
                                elif x_ <= xbc:
                                    tau[j] += np.trapz(dn_dic[:j], x[:j])
                                else:
                                    tau[j] += np.trapz(dp_dic[:j], x[:j])
                                    tau[j] += np.trapz(dn_dic[:j], x[:j]) - np.trapz(
                                        dp_dic[:j], x[:j]
                                    )

                            tau = tau * constants.P_Q

                            tau_e[i_row] = np.trapz(dp_dic[:index_be], x[:index_be]) * constants.P_Q
                            tau_be[i_row] = (
                                np.trapz(dn_dic[:index_be], x[:index_be]) * constants.P_Q
                                - tau_e[i_row]
                            )
                            tau_b[i_row] = (
                                np.trapz(dn_dic[index_be:index_bc], x[index_be:index_bc])
                                * constants.P_Q
                            )
                            tau_c[i_row] = np.trapz(dp_dic[index_bc:], x[index_bc:]) * constants.P_Q
                            tau_bc[i_row] = (
                                np.trapz(dn_dic[index_bc:], x[index_bc:]) * constants.P_Q
                                - tau_c[i_row]
                            )

                            dm_dic = np.where(dn_dic < dp_dic, dn_dic, dn_dic)

                            # for j in range(len(tau)):
                            #     tau[j] = constants.P_Q * np.trapz(dm_dic[:j], x[:j])

                            self.data[key_inqu][specifiers.TRANSIT_TIME] = tau

                        df_iv["tau_e"] = tau_e
                        df_iv["tau_be"] = tau_be
                        df_iv["tau_b"] = tau_b
                        df_iv["tau_bc"] = tau_bc
                        df_iv["tau_c"] = tau_c

                    except:
                        pass

            key_iv = self.join_key(key, "iv")
            self.data[key_iv] = df_iv

            logging.info(
                "Read the Hdev simulation output data of the sweep %s. \nThe simulation folder is %s",
                sweep.name,
                sim_folder,
            )
        except KeyError:
            pass

        # charge partitioning
        # if (
        #     ac and len(dfs_inqu) > 0
        # ):  # only one frequency simulated => post process (maybe also check for 1D)
        #     if not len(freqs) == 2:
        #         raise IOError
        #     # post processing
        #     # 1 general stuff
        #     x = dfs_inqu[0]["X"].to_numpy()
        #     d = dfs_inqu[0][specifiers.NET_DOPING].to_numpy()

        #     junctions = np.where(np.sign(d[:-1]) != np.sign(d[1:]))[0] + 1  # detect sign changes
        #     if junctions != []:  # no junction found
        #         xje = x[junctions[0]]
        #         xjc = x[junctions[1]]

        #         # sum of charge
        #         qp = np.zeros(len(df_iv))
        #         qn = np.zeros(len(df_iv))
        #         for i_row, _row in enumerate(df_iv.iterrows()):
        #             try:
        #                 key_inqu = next(
        #                     _key
        #                     for _key in self.data.keys()
        #                     if "_inqu" in _key and "op" + str(i_row + 1) in _key
        #                 )
        #             except:
        #                 continue
        #             df_inqu = self.data[key_inqu]
        #             qn[i_row] = np.trapz(df_inqu["N"], df_inqu["X"])
        #             qp[i_row] = np.trapz(df_inqu["P"], df_inqu["X"])

        #         df_iv["Q|N"] = qn
        #         df_iv["Q|P"] = qp

        #         # where inqu df is required for every op
        #         gm = np.array(df_iv[specifiers.TRANSCONDUCTANCE])
        #         tau_e = np.ones(len(df_iv))
        #         tau_be = np.ones(len(df_iv))
        #         tau_b = np.ones(len(df_iv))
        #         tau_c = np.ones(len(df_iv))
        #         tau_bc = np.ones(len(df_iv))

        #         try:

        #             for i_row, row in enumerate(df_iv.iterrows()):
        #                 key_inqu = next(
        #                     _key
        #                     for _key in self.data.keys()
        #                     if "_inqu" in _key and "op" + str(i_row + 1) in _key
        #                 )
        #                 key_ac_inqu = next(
        #                     _key
        #                     for _key in self.data.keys()
        #                     if "port1_acinqu" in _key and "op" + str(i_row + 1) in _key
        #                 )

        #                 # get the necessary data
        #                 df_ac_inqu_i = self.data[key_ac_inqu]
        #                 dn_dic = df_ac_inqu_i["re_d_n_x"].to_numpy() / gm[i_row]
        #                 dp_dic = df_ac_inqu_i["re_d_p_x"].to_numpy() / gm[i_row]
        #                 droh_dic = dp_dic - dn_dic
        #                 # transit time
        #                 changes = np.where(np.sign(droh_dic[:-1]) != np.sign(droh_dic[1:]))[0] + 1
        #                 index_be = changes[np.argmin(np.abs(changes - junctions[0]))]
        #                 index_bc = changes[np.argmin(np.abs(changes - junctions[1]))]
        #                 xbe = x[index_be]
        #                 xbc = x[index_bc]

        #                 tau = np.zeros(len(x))
        #                 for j, x_ in enumerate(x):
        #                     if x_ <= xbe:
        #                         tau[j] += np.trapz(dp_dic[:j], x[:j])
        #                         tau[j] += np.trapz(dn_dic[:j], x[:j]) - np.trapz(dp_dic[:j], x[:j])
        #                     elif x_ <= xbc:
        #                         tau[j] += np.trapz(dn_dic[:j], x[:j])
        #                     else:
        #                         tau[j] += np.trapz(dp_dic[:j], x[:j])
        #                         tau[j] += np.trapz(dn_dic[:j], x[:j]) - np.trapz(dp_dic[:j], x[:j])

        #                 tau = tau * constants.P_Q

        #                 tau_e[i_row] = np.trapz(dp_dic[:index_be], x[:index_be]) * constants.P_Q
        #                 tau_be[i_row] = (
        #                     np.trapz(dn_dic[:index_be], x[:index_be]) * constants.P_Q - tau_e[i_row]
        #                 )
        #                 tau_b[i_row] = (
        #                     np.trapz(dn_dic[index_be:index_bc], x[index_be:index_bc])
        #                     * constants.P_Q
        #                 )
        #                 tau_c[i_row] = np.trapz(dp_dic[index_bc:], x[index_bc:]) * constants.P_Q
        #                 tau_bc[i_row] = (
        #                     np.trapz(dn_dic[index_bc:], x[index_bc:]) * constants.P_Q - tau_c[i_row]
        #                 )

        #                 dm_dic = np.where(dn_dic < dp_dic, dn_dic, dn_dic)

        #                 for j in range(len(tau)):
        #                     tau[j] = constants.P_Q * np.trapz(dm_dic[:j], x[:j])

        #                 self.data[key_inqu][specifiers.TRANSIT_TIME] = tau

        #             df_iv["tau_e"] = tau_e
        #             df_iv["tau_be"] = tau_be
        #             df_iv["tau_b"] = tau_b
        #             df_iv["tau_bc"] = tau_bc
        #             df_iv["tau_c"] = tau_c

        #         except:
        #             pass

        key_iv = self.join_key(key, "iv")
        self.data[key_iv] = df_iv
        pd.options.mode.chained_assignment = "warn"  # default='warn'

        logging.info(
            "Read the Hdev simulation output data of the sweep %s. \nThe simulation folder is %s",
            sweep.name,
            sim_folder,
        )
        if delete_sim_results:
            self.delete_sim_results(sweep)

    def get_bias_info(self, sweep, sub_sweep):
        # get contact
        nodes = sub_sweep.var_name.nodes
        if len(nodes) > 1:
            raise IOError("DMT -> DutHdev: currently only potential sweeps are supported.")

        try:
            cont_name = "'" + nodes[0] + "'"
        except IndexError:
            cont_name = "'" + str(sub_sweep.var_name)[0] + "'"

        if sub_sweep.sweep_type == "LIN":
            bias_fun = "'LIN'"
            bias_info = {
                "cont_name": cont_name,
                "bias_fun": bias_fun,
            }
        elif sub_sweep.sweep_type.startswith("CON"):
            bias_fun = "'TAB'"  # we convert to list, to support list=1 setting used by DMT in Hdev
            bias_info = {
                "cont_name": cont_name,
                "bias_fun": bias_fun,
                "bias_val": [sub_sweep.value_def[0]],
            }
        elif sub_sweep.sweep_type.startswith("SYNC"):
            # find master sweep
            for other_sub_sweep in sweep.sweepdef:
                if other_sub_sweep.var_name == sub_sweep.master_var:
                    bias_info = self.get_bias_info(sweep, other_sub_sweep)
                    bias_info["cont_name"] = cont_name
                    if bias_info["bias_fun"] == "'LIN'":
                        bias_info["bias_val"][0] = bias_info["bias_val"][0] + sub_sweep.offset
                        bias_info["bias_val"][1] = bias_info["bias_val"][1] + sub_sweep.offset
                    else:
                        raise NotImplementedError
                    dummy = 1
                    break

        elif sub_sweep.sweep_type.startswith("LIST"):
            bias_fun = "'TAB'"  # we convert to list, to support list=1 setting used by DMT in Hdev
            bias_info = {"cont_name": cont_name, "bias_fun": bias_fun, "bias_val": sub_sweep.values}
        else:
            raise NotImplementedError

        return bias_info  # type: ignore

    def get_mobility(self, semiconductor, valley, field, temperature, doping, grading):
        if not DutHdev.inited:
            self.init_()

        if type(field) == int:
            mob = np.zeros_like(doping)
            for i in range(len(doping)):
                if doping[i] > 0:
                    mob[i] = hdev_py.get_mobility_py(
                        semiconductor,
                        valley,
                        float(field),
                        1,
                        1,
                        1,
                        temperature,
                        0,
                        doping[i],
                        0,
                        grading,
                    )
                    # def get_mobility_py(semi_name     ,valley,f       , ec_l, ec_r, dim, t, dens, don, acc,  grad):
                else:
                    mob[i] = hdev_py.get_mobility_py(
                        semiconductor,
                        valley,
                        float(field),
                        1,
                        1,
                        1,
                        temperature,
                        0,
                        0,
                        -doping[i],
                        grading,
                    )
        else:
            mob = np.zeros_like(field)
            for i in range(len(field)):
                if doping > 0:
                    mob[i] = hdev_py.get_mobility_py(
                        semiconductor, valley, field[i], 1, 1, 1, temperature, 0, doping, 0, grading
                    )
                    # def get_mobility_py(semi_name     ,valley,f       , ec_l, ec_r, dim, t, dens, don, acc,  grad):
                else:
                    mob[i] = hdev_py.get_mobility_py(
                        semiconductor,
                        valley,
                        field[i],
                        1,
                        1,
                        1,
                        temperature,
                        0,
                        0,
                        -doping,
                        grading,
                    )

        return mob

    def get_intervalley_rate(self, semiconductor, valley_1, valley_2, field, temperature, doping):
        if not DutHdev.inited:
            self.init_()

        rec = np.zeros_like(field)
        for i in range(len(field)):
            rec[i] = hdev_py.get_intervalley_rate_py(
                semiconductor, valley_1, valley_2, field[i], doping, temperature
            )

        return rec

    def get_mobility_paras(self, semi, valley):
        if not DutHdev.inited:
            self.init_()

        paras = hdev_py.get_mobility_paras_py(semi, valley)
        return paras

    def get_recombination_paras(self, semi, rec_type):
        if not DutHdev.inited:
            self.init_()

        paras = hdev_py.get_recombination_paras_py(semi, rec_type)
        return paras

    def set_mobility_paras(self, semi, valley, paras):
        if not DutHdev.inited:
            self.init_()

        hdev_py.set_mobility_paras_py(semi, valley, paras)
        return

    def set_recombination_paras(self, semi, paras):
        if not DutHdev.inited:
            self.init_()

        hdev_py.set_recombination_paras_py(semi, paras)
        return

    def init_(self):
        DutHdev.inited = True
        # load structure
        inp_str = makeinp(self._inp_structure)
        # load sweep
        # create a sweep
        sweepdef = []
        n = 1
        for region_def in self._inp_structure["REGION_DEF"]:
            if "cont_name" in region_def.keys():
                sweepdef.append(
                    {
                        "var_name": specifiers.VOLTAGE + region_def["cont_name"],
                        "sweep_order": n,
                        "sweep_type": "CON",
                        "value_def": [0],
                    }
                )
                n = n + 1
        outputdef = []
        othervar = {"TEMP": 300}
        sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)
        inp_str = self.make_input(sweep)

        path_sim = os.path.join(DATA_CONFIG["directories"]["simulation"], "inp_tmp.inp")
        inp_file = open(path_sim, "w+")
        inp_file.write(inp_str)
        inp_file.close()
        shutil.copyfile(
            os.path.join(DATA_CONFIG["directories"]["simulation"], "bias.h5"),
            os.path.join("/home/markus/Documents/Gitprojects/dissertation/bias.h5"),
        )
        hdev_py.init(inp_file.name)

    def scale_modelcard(self, *_args, **_kwargs):
        """Hdev currently has no scaling"""
        pass


def getITRSHBT(
    hbt_fun=None,
    fermi=False,
    nl=False,
    sat=False,
    tn=True,
    no_inqu=False,
    lambda_e=None,
    xj_l=None,
    mu_min_a=None,
    mu_min_d=None,
    beta=None,
):
    """Return a DutHdev for ITRS HBTS. Only for SiGe HBTs.

    Parameters
    ----------
    hbt_fun : callable
        Function that returns a Hdev HBT definition.
    fermi : bool
        If True: use Fermi-Dirac statistics, else Boltzmann
    nl : bool
        If True: use Non local energy model.
    sat : bool
        If True: use velocity saturation model.
    tn : bool
        If True: use ET model, else DD model.

    Returns
    -------
    dut_hdev : DutHdev
        A DMT.Hdev.DutHdev object ready for simulation.
    """
    if hbt_fun is None:
        raise IOError("You need to provide a function that returns a valid Hdev HBT definition.")

    inp = hbt_fun()

    if no_inqu:
        inp["OUTPUT"]["inqu_lev"] = 0

    if fermi:
        for i, semi in enumerate(inp["SEMI"]):
            inp["SEMI"][i]["fermi"] = 1

    if nl:  # todo: should depend on profile
        inp["NON_LOCAL"] = {}
        inp["NON_LOCAL"]["type"] = "simple"
        inp["NON_LOCAL"]["lambda"] = lambda_e
        inp["NON_LOCAL"]["xj_l"] = xj_l
        # increase velocity sat.
        # for l, mob in enumerate(inp["MOB_DEF"]):
        #     if inp["MOB_DEF"][l]["valley"] == "X":
        #         inp["MOB_DEF"][l]["v_sat"] = inp["MOB_DEF"][l]["v_sat"] * 1.5
    if tn:
        inp["DD"]["tn"] = 1

    if not sat:
        for i, _mob in enumerate(inp["MOB_DEF"]):
            inp["MOB_DEF"][i]["hc_scat_type"] = "default"
    else:
        for i, _mob in enumerate(inp["MOB_DEF"]):
            if inp["MOB_DEF"][i]["valley"] == "x":
                inp["MOB_DEF"][i]["hc_scat_type"] = "sat"

    if mu_min_a is not None:
        # decrease lattice mobility
        for l, mob in enumerate(inp["MOB_DEF"]):
            if inp["MOB_DEF"][l]["valley"] == "X":
                inp["MOB_DEF"][l]["mu_min_a"] = inp["MOB_DEF"][l]["mu_min_a"] * mu_min_a

    if mu_min_d is not None:
        # decrease lattice mobility
        for l, mob in enumerate(inp["MOB_DEF"]):
            if inp["MOB_DEF"][l]["valley"] == "X":
                inp["MOB_DEF"][l]["mu_min_d"] = inp["MOB_DEF"][l]["mu_min_d"] * mu_min_d

    if beta is not None:
        # decrease lattice mobility
        for l, mob in enumerate(inp["MOB_DEF"]):
            if inp["MOB_DEF"][l]["valley"] == "X":
                inp["MOB_DEF"][l]["beta"] = 1.5

    inp = turn_off_recombination(inp)

    return DutHdev(
        None,
        DutType.npn,
        inp,
        reference_node="E",
    )



""" DMT config management

Collects the config from 3 different locations:

* Local script directory (DMT_config.yaml).
* User's home directory (%LOCALAPPDATA%\DMT\DMT_config.yaml or $XDG_CONFIG_HOME/DMT/DMT_config.yaml with $XDG_CONFIG_HOME defaulting to ~/.config) and
* DMT package installation directory (DMT/config/DMT_config.yaml)

They are all read and finally taken in the order given here. This means that anything given in the local directory overwrites all others.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

name = "config"

import os
import platform
import pkgutil
import warnings
import yaml
from pathlib import Path

# dmt default config
path_config = Path(__file__).resolve().parent
default_config_file = path_config / "DMT_config.yaml"
with open(default_config_file) as yaml_data_file:
    DATA_CONFIG = yaml.safe_load(yaml_data_file)

# dmt user config
data_user = {}

if platform.system() == "Windows":
    user_config = Path(os.getenv("LOCALAPPDATA")).expanduser()
elif platform.system() == "Linux":
    user_config = Path(
        os.getenv("XDG_CONFIG_HOME") if os.getenv("XDG_CONFIG_HOME", False) else "~/.config"
    ).expanduser()
elif platform.system() == "Darwin":
    user_config = Path(
        os.getenv("XDG_CONFIG_HOME") if os.getenv("XDG_CONFIG_HOME", False) else "~/.config"
    ).expanduser()
else:
    raise OSError(
        f"Unknown platform:{platform.system()}. Currently only Windows, Linux and MacOS are recognized"
    )

user_config = user_config / "DMT" / "DMT_config.yaml"
try:
    with user_config.open() as yaml_data_file:
        data_user = yaml.safe_load(yaml_data_file)
except FileNotFoundError:
    try:
        user_config = Path.home() / ".DMT" / "DMT_config.yaml"
        with user_config.open() as yaml_data_file:
            data_user = yaml.safe_load(yaml_data_file)
        # warnings.warn(
        #     "The DMT user configuration file is moved. The new paths are:",
        #     "Windows: %LOCALAPPDATA%\DMT\DMT_config.yaml",
        #     "Linux and MacOS: $XDG_CONFIG_HOME/DMT/DMT_config.yaml, defaulting to ~/.config/DMT/DMT_config.yaml",
        #     category=DeprecationWarning,
        # )
    except FileNotFoundError:
        pass

if data_user:
    for key, value in DATA_CONFIG.items():
        if key in data_user.keys():
            if hasattr(DATA_CONFIG[key], "update"):
                DATA_CONFIG[key].update(data_user[key])
            else:
                if isinstance(DATA_CONFIG[key], list):  # lists are appended.
                    DATA_CONFIG[key] = DATA_CONFIG[key] + data_user[key]
                else:
                    DATA_CONFIG[key] = data_user[key]

# finally workspace path
try:
    local_config = Path("DMT_config.yaml")
    with local_config.open() as yaml_data_file:
        data_folder = yaml.safe_load(yaml_data_file)

    for key, value in DATA_CONFIG.items():
        if key in data_folder.keys():
            try:
                DATA_CONFIG[key].update(data_folder[key])
            except AttributeError:
                if isinstance(DATA_CONFIG[key], list):  # lists are appended.
                    DATA_CONFIG[key] = DATA_CONFIG[key] + data_folder[key]
                else:
                    DATA_CONFIG[key] = data_folder[key]
except FileNotFoundError:
    pass

for key, str_path in DATA_CONFIG["directories"].items():
    try:
        DATA_CONFIG["directories"][key] = Path(str_path).expanduser().resolve()
    except TypeError:
        pass


# measurement data overview report template
if DATA_CONFIG["directories"]["libautodoc"] is None:
    DATA_CONFIG["directories"]["libautodoc"] = path_config.parent / "libautodoc_template"
else:
    DATA_CONFIG["directories"]["libautodoc"] = (
        Path(DATA_CONFIG["directories"]["libautodoc"]).expanduser().resolve()
    )

# xtraction result overview report template
if DATA_CONFIG["directories"]["autodoc"] is None:
    package = pkgutil.get_loader("DMT.extraction")
    if package is not None:
        path_autodoc_default = (
            Path(package.get_filename()).resolve().parent.parent / "autodoc_template"
        )
        DATA_CONFIG["directories"]["autodoc"] = Path(path_autodoc_default)
else:
    DATA_CONFIG["directories"]["autodoc"] = Path(DATA_CONFIG["directories"]["autodoc"]).expanduser()

# not needed anymore. It is directly obtained from DATA_CONFIG.
# This allows to add configs without changing anything here. Do it always like this in future!
COMMANDS = DATA_CONFIG["commands"]

COMMAND_TEX = (DATA_CONFIG["commands"]["TEX"], DATA_CONFIG["commands"]["TEXARGS"])
""" TEX build command """

# DO NOT ADD ANYTHING HERE. JUST DIRECTLY IMPORT DATA_CONFIG INSIDE YOUR MODULE
# If wanted add documentation here and/or in DMT_config.yaml


def print_config(file_path=None):
    """Prints the current used config either to stdout or into a file

    Parameters
    ----------
    file_path : str, optional
        Path to print the config to, by default None.
    """
    if file_path is None:
        print(DATA_CONFIG)
    else:
        yaml.safe_dump(DATA_CONFIG, file_path)


from DMT.external.os import recursive_copy


def generate_libdoc_template(directory):
    """Copies the DMT template for the DutLib documentation into the given directory. This allows to change the template according to your needs.

    Parameters
    ----------
    directory : str
        Path to copy to.
    """
    if isinstance(directory, Path):
        destination = directory
    else:
        destination = Path(directory)

    destination.mkdir(parents=True, exist_ok=True)
    recursive_copy(DATA_CONFIG["directories"]["libautodoc"], destination)


def export_autodoc_template(directory):
    """Copies the DMT template for the extraction documentation into the given directory. This allows to change the template according to your needs.

    Only possible if the extraction module is installed!

    Parameters
    ----------
    directory : str
        Path to copy to.
    """
    if isinstance(directory, Path):
        destination = directory
    else:
        destination = Path(directory)

    destination.mkdir(parents=True, exist_ok=True)
    recursive_copy(DATA_CONFIG["directories"]["autodoc"], destination)



import pytest
import pandas as pd
import numpy as np
from DMT.core import Sweep, specifiers
from DMT.core.sweep_def import (
    SweepDef,
    SweepDefConst,
    SweepDefLinear,
    SweepDefLog,
    SweepDefSync,
    SweepDefList,
)

sp_temp = specifiers.TEMPERATURE
sp_freq = specifiers.FREQUENCY
sp_ft = specifiers.TRANSIT_FREQUENCY
sp_vb = specifiers.VOLTAGE + "B"
sp_vcb = specifiers.VOLTAGE + ["C", "B"]
sp_vbc = specifiers.VOLTAGE + ["B", "C"]
sp_vc = specifiers.VOLTAGE + "C"
sp_ve = specifiers.VOLTAGE + "E"
sp_ic = specifiers.CURRENT + "C"
sp_ib = specifiers.CURRENT + "B"


def test_dc_sweep():
    # sweeps are defined using a sweep_def
    sweepdef = [
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 1,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [1],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 3,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    # additional variables that shall be saved in the sweep
    othervar = {"TEMP": 300, "w": 10, "l": 0.25}

    # Sweep as a class, inputs are needed, e.g.
    # name:     dummy-sweep
    # sweepdef: sweepdef defined above related opering points
    # othervar: othervar defined above related T, geometries
    sweep = Sweep("dummy-sweep", sweepdef=sweepdef, othervar=othervar)  # create the Sweep object
    df = (
        sweep.create_df()
    )  # we can also create the sweep's dataframe, where the output variables are Nans.

    df.ensure_specifier_column(sp_vbc)
    assert len(df) == 11
    assert df[sp_vc].round(3).unique() == np.array([1])
    assert np.allclose(df[sp_vb], np.linspace(0, 1, 11))


def test_sync_sweep():

    # VBC = 0.1
    sweepdef = [
        {
            "var_name": sp_vb,
            "sweep_order": 1,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": sp_vc,
            "sweep_order": 1,
            "sweep_type": "SYNC",
            "master": sp_vb,
            "offset": 0.1,
        },
        {
            "var_name": sp_ve,
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = ["I_C", "I_B"]  # define which output variables need to be saved
    othervar = {"TEMP": 300, "w": 10, "l": 0.25}
    sweep = Sweep("dummy_2", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)
    df = sweep.create_df()

    df.ensure_specifier_column(sp_vbc)
    assert len(df) == 11
    assert df[sp_vbc].round(3).unique() == np.array([-0.1])
    assert np.allclose(df[sp_vb], np.linspace(0, 1, 11))
    assert np.allclose(df[sp_vc], np.linspace(0.1, 1.1, 11))

    list_vbc = [0.5, 0.0, -0.5]
    sws_gummel = []
    df_gummel_sep = []
    for vbc in list_vbc:
        sws_gummel.append(
            Sweep(
                f"gummel_vbc_{vbc:.1f}",
                sweepdef=[
                    SweepDefConst(sp_ve, [0], sweep_order=0),
                    SweepDefLinear(sp_vb, [0.8, 1.0, 41], sweep_order=1),
                    SweepDefSync(sp_vc, sp_vb, vbc, sweep_order=1),
                    SweepDefLog(sp_freq, [9, 10, 5], sweep_order=2),
                ],
                outputdef=[],
                othervar={sp_temp: 300},
            )
        )
        df_gummel_sep.append(sws_gummel[-1].create_df())

    df_gummel_sep = pd.concat(df_gummel_sep)

    sw_gummel = Sweep(
        "gummel",
        sweepdef=[
            SweepDefConst(sp_ve, 0, sweep_order=0),
            SweepDefList(sp_vbc, list_vbc, sweep_order=2),
            SweepDefLinear(sp_vb, [0.8, 1.0, 41], sweep_order=2),
            SweepDefSync(sp_vc, sp_vb, sp_vbc, sweep_order=2),
            SweepDefLog(sp_freq, [9, 10, 5], sweep_order=3),
        ],
        outputdef=[],
        othervar={sp_temp: 300},
    )
    df_gummel = sw_gummel.create_df()
    df_gummel.drop(columns=sp_vbc, inplace=True)

    # compare the dataframes -> should be equal
    for row_gummel, row_gummel_sep in zip(df_gummel.iterrows(), df_gummel_sep.iterrows()):
        assert all(row_gummel[1] == row_gummel_sep[1])


def test_ac_sweep():
    # create a sweep with a sweeped offset variable
    sweepdef = [
        {"var_name": sp_freq, "sweep_order": 4, "sweep_type": "LOG", "value_def": [8, 9, 2]},
        {
            "var_name": sp_vb,
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0, 1, 3],
        },
        {
            "var_name": sp_vc,
            "sweep_order": 3,
            "sweep_type": "SYNC",
            "master": sp_vb,
            "offset": sp_vcb,
        },
        {
            "var_name": sp_vcb,
            "sweep_order": 2,
            "sweep_type": "LIST",
            "value_def": [-0, -0.2, -0.3],
        },
        {
            "var_name": sp_ve,
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = [specifiers.CURRENT + "C", specifiers.CURRENT + "B"]
    othervar = {"TEMP": 300, "w": 10, "l": 0.25}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)
    df = sweep.create_df()

    df.ensure_specifier_column(sp_vbc)
    assert len(df) == 18
    assert np.allclose(df[sp_vbc].round(3).unique(), np.array([0.0, 0.2, 0.3]))
    assert np.allclose(df[sp_vb].round(3).unique(), np.array([0.0, 0.5, 1.0]))
    assert np.allclose(
        df[sp_vc].round(3).unique(), np.array([0.0, 0.5, 1.0, -0.2, 0.3, 0.8, -0.3, 0.2, 0.7])
    )


def test_sweepdef_errors():
    with pytest.raises(IOError):
        SweepDef("asd", "LIN")  # var_name 'asd' can not be converted to specifier

    with pytest.raises(IOError):
        SweepDef("FREQ", "LIN")  # neither master not value_def

    with pytest.raises(IOError):
        SweepDef("FREQ", "LON")  # typo in sweep type

    with pytest.raises(IOError):
        swd = SweepDef("FREQ", "CON", value_def=[0, 1])  # 2 values for constant
        swd.set_values()

    swd = SweepDef("FREQ", "LOG", value_def=[8, 9, 2])
    assert swd.sync is None


def test_sweep_swd():

    swd = SweepDef("FREQ", "LOG", value_def=[8, 9, 2])
    swds = [
        swd,
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 1,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [1],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 3,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    with pytest.raises(IOError):  # no temperature
        Sweep("gummel", sweepdef=swds)

    with pytest.raises(IOError):  # 2 temperatures
        Sweep(
            "gummel",
            sweepdef=[
                {"var_name": "TEMP", "sweep_order": 0, "sweep_type": "CON", "value_def": [300]},
            ]
            + swds,
            othervar={"TEMP": 300},
        )

    with pytest.raises(IOError):  # othervar must be a dict
        Sweep("gummel", sweepdef=swds, othervar="TEMP")

    with pytest.raises(IOError):  # outputdef must be a list
        Sweep("gummel", sweepdef=swds, othervar={"TEMP": 300}, outputdef="I_C")

    with pytest.raises(IOError):  # no master sweep
        Sweep(
            "gummel",
            sweepdef=[
                {
                    "var_name": specifiers.VOLTAGE + "B",
                    "sweep_order": 1,
                    "sweep_type": "LIN",
                    "value_def": [0, 1, 11],
                },
                {
                    "var_name": specifiers.VOLTAGE + "C",
                    "sweep_order": 1,
                    "sweep_type": "SYNC",
                    "master": "V_A",
                    "offset": 0.1,
                },
                {
                    "var_name": specifiers.VOLTAGE + "E",
                    "sweep_order": 2,
                    "sweep_type": "CON",
                    "value_def": [0],
                },
            ],
        )

    with pytest.raises(IOError):  # no master sweep
        Sweep(
            "gummel",
            sweepdef=[
                {
                    "var_name": specifiers.VOLTAGE + "B",
                    "sweep_type": "LIN",
                    "value_def": [0, 1, 11],
                },
                {
                    "var_name": specifiers.VOLTAGE + "C",
                    "sweep_type": "SYNC",
                    "master": "V_A",
                    "offset": 0.1,
                },
                {"var_name": specifiers.VOLTAGE + "E", "sweep_type": "CON", "value_def": [0]},
            ],
        )

    sweep = Sweep("gummel", sweepdef=swds, othervar={"TEMP": 300})
    sweep2 = Sweep("gummel", sweepdef=swds, othervar={"TEMP": 300})

    assert sweep == sweep2

    assert sweep != 5


def test_sweep_temp():

    assert Sweep("gummel", sweepdef=[], othervar={"TEMP": 300}).get_temperature() == "T300.00K"

    assert (
        Sweep(
            "gummel",
            sweepdef=[
                {"var_name": "TEMP", "sweep_type": "CON", "value_def": [300]},
            ],
        ).get_temperature()
        == "T300.00K"
    )

    assert (
        Sweep(
            "gummel",
            sweepdef=[
                {"var_name": "TEMP", "sweep_type": "LIN", "value_def": [300, 400, 3]},
            ],
        ).get_temperature()
        == "T[300.00-3.00-400.00]K"
    )

    assert (
        Sweep(
            "gummel",
            sweepdef=[
                {"var_name": "TEMP", "sweep_type": "LIST", "value_def": [300, 320, 350]},
            ],
        ).get_temperature()
        == "T(300.00,320.00,350.00)K"
    )

    with pytest.raises(NotImplementedError):  # log sweep has not string at the moment
        Sweep(
            "gummel",
            sweepdef=[
                {"var_name": "TEMP", "sweep_type": "LOG", "value_def": [2, 4, 10]},
            ],
        ).get_temperature()


if __name__ == "__main__":
    test_dc_sweep()
    test_sync_sweep()
    test_ac_sweep()
    test_sweepdef_errors()
    test_sweep_swd()
    test_sweep_temp()



import os
import pandas as pd
from DMT.core import Plot, specifiers
import numpy as np

if __name__ == "__main__":
    # open file and read
    path = os.path.join("test/test_interface_ngspice/output_tran")
    with open(path, "r") as f:
        txt = f.read()

    txt = txt.splitlines()
    headers = txt[0].split()
    data = {}
    for head in headers:
        data[head] = np.zeros(len(txt) - 1)

    for i, line in enumerate(txt[1:]):
        line = line.split()
        for head, ele in zip(headers, line):
            data[head][i] = float(ele)

    data = pd.DataFrame.from_dict(data)

    data_ads = pd.read_csv("test/test_interface_ngspice/transient_data_ads.csv")

    plt_cc = Plot(
        "V_C(t)",
        style="xtraction_color",
        x_label="t(ps)",
        y_specifier=specifiers.VOLTAGE + "C",
        x_scale=1e12,
    )
    plt_cc.add_data_set(data["time"], data["c"], label="ngs")
    plt_cc.add_data_set(data_ads["t"], data_ads["V_C"], label="ads")
    plt_cc.plot_py(show=True)



""" DMT's external tools.

These are routines and classes which manage the interfaces to other Python packages. For example PyLaTeX.
"""
from .os import cd, recursive_copy, slugify, rmtree

## build tex
from .latex import (
    build_tex,
    build_svg,
    clean_tex_files,
    build_png,
    SI_UNITS_CONVERTER,
    resolve_siunitx,
    tex_to_text,
)

## pylatex classes
try:
    from .pylatex import Tex
    from .pylatex import SubFile
    from .pylatex import CommandBase
    from .pylatex import CommandInput
    from .pylatex import CommandInputTikz
    from .pylatex import CommandLabel
    from .pylatex import CommandRef
    from .pylatex import CommandRefRange
except ImportError:
    pass

## verilogae
from .verilogae import HICUM_L0
from .verilogae import HICUM_L2
from .verilogae import SGP

from .verilogae import get_param_list
from .verilogae import get_dmt_model



import numpy as np
from pathlib import Path
from DMT.core import read_data, DataFrame, specifiers, sub_specifiers


def test_clean_data():
    data = read_data(Path(__file__).resolve().parent / "test_data" / "dummy_open_freq.mdm")
    data_clean = data.clean_data(nodes=["B", "E", "C"], reference_node="E", ac_ports=["B", "C"])
    return data_clean


def test_clean_data_potentials():
    # create a simple df that has only voltages
    df = DataFrame()
    vbe = np.linspace(0, 1, 101)
    vce = np.linspace(0, 1, 101)
    df[specifiers.VOLTAGE + "B" + "E"] = vbe
    df[specifiers.VOLTAGE + "C" + "E"] = vce
    df = df.clean_data(nodes=["B", "E", "C"], reference_node="E")
    print(df.columns)
    assert specifiers.VOLTAGE + "B" in df.columns
    assert specifiers.VOLTAGE + "C" in df.columns
    assert specifiers.VOLTAGE + "E" in df.columns
    assert specifiers.VOLTAGE + "B" + "E" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" not in df.columns


def test_clean_data_potential_mix():
    # create a simple df that has only voltages
    df = DataFrame()
    vbe = np.linspace(0, 1, 101)
    vce = np.linspace(0, 1, 101)
    df[specifiers.VOLTAGE + "B" + "E"] = vbe
    df[specifiers.VOLTAGE + "C"] = vce
    df = df.clean_data(nodes=["B", "E", "C"], reference_node="E")
    print(df.columns)
    assert specifiers.VOLTAGE + "B" in df.columns
    assert specifiers.VOLTAGE + "C" in df.columns
    assert specifiers.VOLTAGE + "E" in df.columns
    assert specifiers.VOLTAGE + "B" + "E" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" not in df.columns


def test_clean_data_potential_mix_2():
    # create a simple df that has only voltages
    df = DataFrame()
    vbe = np.linspace(0, 1, 101)
    vce = np.linspace(0, 1, 101)
    df[specifiers.VOLTAGE + "B"] = vbe
    df[specifiers.VOLTAGE + "C"] = vce
    df = df.clean_data(nodes=["B", "E", "C"], reference_node="E")
    print(df.columns)
    assert specifiers.VOLTAGE + "B" in df.columns
    assert specifiers.VOLTAGE + "C" in df.columns
    assert specifiers.VOLTAGE + "E" in df.columns
    assert specifiers.VOLTAGE + "B" + "E" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" not in df.columns


def test_clean_data_potential_mix_3():
    # create a simple df that has only voltages
    df = DataFrame()
    vbe = np.linspace(0, 1, 101)
    vce = np.linspace(0, 1, 101)
    df[specifiers.VOLTAGE + "B"] = vbe
    df[specifiers.VOLTAGE + "B" + "C"] = vce
    df = df.clean_data(nodes=["B", "E", "C"], reference_node="E")
    print(df.columns)
    assert specifiers.VOLTAGE + "B" in df.columns
    assert specifiers.VOLTAGE + "C" in df.columns
    assert specifiers.VOLTAGE + "E" in df.columns
    assert specifiers.VOLTAGE + "B" + "E" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" not in df.columns


def test_clean_data_throw_error():
    # create a simple df that has only voltages
    df = DataFrame()
    vbe = np.linspace(0, 1, 101)
    vce = np.linspace(0, 1, 101)
    df[specifiers.VOLTAGE + "B" + "E" + "FORCED"] = vbe
    df[specifiers.VOLTAGE + "B" + "E"] = vbe
    df[specifiers.VOLTAGE + "C" + "E" + "FORCED"] = vbe
    df[specifiers.VOLTAGE + "C" + "E"] = vbe
    df = df.clean_data(nodes=["B", "E", "C"], reference_node="E")
    print(df.columns)
    assert specifiers.VOLTAGE + "B" in df.columns
    assert specifiers.VOLTAGE + "C" in df.columns
    assert specifiers.VOLTAGE + "E" in df.columns
    assert specifiers.VOLTAGE + "B" + "E" not in df.columns
    assert specifiers.VOLTAGE + "B" + "E" + "FORCED" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" not in df.columns
    assert specifiers.VOLTAGE + "C" + "E" + "FORCED" not in df.columns


if __name__ == "__main__":
    data_clean = test_clean_data()
    test_clean_data_potentials()
    test_clean_data_potential_mix()
    test_clean_data_potential_mix_2()
    test_clean_data_potential_mix_3()
    test_clean_data_throw_error()

    # print(data_clean['S_BB'].to_numpy())



""" Python script which is explained in detail in the introduction.
Here the commands are repeated to be able to test them.
"""
from pathlib import Path
from DMT import core


path_data = Path(__file__).resolve().parent.parent / "_static" / "intro"

dut_meas = core.DutMeas(
    database_dir=None,  # Use dir from config
    dut_type=core.DutType.npn,  # it is a BJT-npn
    width=float(1e-6),
    length=float(2e-6),
    name="dut_meas_npn",  # name and width/length for documentation
    reference_node="E",  # defines configuration
)
key_saved = "300K/iv"
dut_meas.add_data(path_data / "meas_data_300K.csv", key=key_saved)
dut_meas.clean_data(fallback={"E_": "E"})

print(dut_meas.data[key_saved].columns)
# V_c, V_E, V_B, I_E_, I_C, I_b, FREQ, Y_21, Y_11, Y22, Y_12

modelcard = core.MCard(
    ["c", "b", "e", "s"],
    "QSGP1",
    core.circuit.SGP_BJT,
    1.0,
    va_file=path_data / "sgp_v1p0.va",
)
modelcard.load_model_parameters(path_data / "bjt.lib")


def get_circuit(self):
    """Returns a circuit which uses the modelcard to which the method is attached.

    Returns
    -------
    circuit : :class:`~DMT.core.circuit.Circuit`

    """
    circuit_elements = []
    # model instance
    circuit_elements.append(
        core.circuit.CircuitElement(
            self.default_module_name,
            self.default_subckt_name,
            [f"n_{node.upper()}" for node in self.nodes_list],
            # ["n_C", "n_B", "n_E"],
            parameters=self,
        )
    )

    # BASE NODE CONNECTION #############
    # shorts for current measurement
    circuit_elements.append(
        core.circuit.CircuitElement(core.circuit.SHORT, "I_B", ["n_B_FORCED", "n_B"])
    )
    # COLLECTOR NODE CONNECTION #############
    circuit_elements.append(
        core.circuit.CircuitElement(core.circuit.SHORT, "I_C", ["n_C_FORCED", "n_C"])
    )
    # EMITTER NODE CONNECTION #############
    circuit_elements.append(
        core.circuit.CircuitElement(core.circuit.SHORT, "I_E", ["n_E_FORCED", "n_E"])
    )
    # add sources
    circuit_elements.append(
        core.circuit.CircuitElement(
            core.circuit.VOLTAGE,
            "V_B",
            ["n_B_FORCED", "0"],
            parameters=[("Vdc", "V_B"), ("Vac", "V_B_ac")],
        )
    )
    circuit_elements.append(
        core.circuit.CircuitElement(
            core.circuit.VOLTAGE,
            "V_C",
            ["n_C_FORCED", "0"],
            parameters=[("Vdc", "V_C"), ("Vac", "V_C_ac")],
        )
    )
    circuit_elements.append(
        core.circuit.CircuitElement(
            core.circuit.VOLTAGE,
            "V_E",
            ["n_E_FORCED", "0"],
            parameters=[("Vdc", "V_E"), ("Vac", "V_E_ac")],
        )
    )

    # metal resistance between contact emitter potential and substrate contact
    circuit_elements.append(
        core.circuit.CircuitElement(
            core.circuit.RESISTANCE, "R_S", ["n_S", "n_E_FORCED"], parameters=[("R", str(0.5))]
        )
    )

    # some variables used in this circuit
    circuit_elements += [
        "V_B=0",
        "V_C=0",
        "V_E=0",
        "ac_switch=0",
        "V_B_ac=1-ac_switch",
        "V_C_ac=ac_switch",
        "V_E_ac=0",
    ]

    return core.circuit.Circuit(circuit_elements)


import types

modelcard.get_circuit = types.MethodType(get_circuit, modelcard)


from DMT.xyce import DutXyce

dut_sim = DutXyce(
    None,
    core.DutType.npn,
    modelcard,
    nodes="B,C,E",
    reference_node="E",
)

# from DMT.ngspice import DutNgspice

# dut_sim = DutNgspice(
#     None,
#     core.DutType.npn,
#     modelcard,
#     nodes="B,C,E",
#     reference_node="E",
# )


sweep = core.Sweep.get_sweep_from_dataframe(dut_meas.data[key_saved], temperature=300)


sim_con = core.SimCon()
sim_con.append_simulation(dut=dut_sim, sweep=sweep)
sim_con.run_and_read()

data_meas = dut_meas.data[key_saved]
data_sim = dut_sim.get_data(sweep=sweep)

col_vbe = core.specifiers.VOLTAGE + ["B", "E"]
col_vbc = core.specifiers.VOLTAGE + ["B", "C"]
col_ic = core.specifiers.CURRENT + "C"
col_freq = core.specifiers.FREQUENCY
col_ft = core.specifiers.TRANSIT_FREQUENCY
col_y21_real = core.specifiers.SS_PARA_Y + ["C", "B"] + core.sub_specifiers.REAL

for dut, data in zip([dut_meas, dut_sim], [data_meas, data_sim]):
    data.ensure_specifier_column(col_vbe)
    data.ensure_specifier_column(col_vbc)
    data.ensure_specifier_column(col_ft, ports=dut.ac_ports)
    data.ensure_specifier_column(col_y21_real, ports=dut.ac_ports)

# Plot and save as pdf
plt_ic = core.Plot(
    plot_name="I_C(V_BE)",
    x_specifier=col_vbe,
    y_specifier=col_ic,
    y_scale=1e3,
    y_log=True,
    legend_location="lower right",
)
plt_y21 = core.Plot(
    plot_name="Y_21(I_C)",
    x_specifier=col_ic,
    x_scale=1e3,
    x_log=True,
    y_specifier=col_y21_real,
    y_scale=1e3,
    y_log=True,
    legend_location="lower right",
)
plt_ft = core.Plot(
    plot_name="F_T(I_C)",
    x_specifier=col_ic,
    x_scale=1e3,
    x_log=True,
    y_specifier=col_ft,
    legend_location="upper left",
)

import numpy as np

for source, data in zip(["meas", "sim"], [data_meas, data_sim]):
    for i_vbc, vbc, data_vbc in data.iter_unique_col(col_vbc, decimals=3):
        data_freq = data_vbc[np.isclose(data_vbc[col_freq], 1e7)]
        plt_ic.add_data_set(
            data_freq[col_vbe],
            data_freq[col_ic],
            label=source + " " + col_vbc.to_legend_with_value(vbc),
        )
        plt_y21.add_data_set(
            data_freq[col_ic],
            data_freq[col_y21_real],
            label=source + " " + col_vbc.to_legend_with_value(vbc),
        )
        plt_ft.add_data_set(
            data_freq[col_ic],
            data_freq[col_ft],
            label=source + " " + col_vbc.to_legend_with_value(vbc),
        )

plt_ic.plot_pyqtgraph(show=False)
plt_y21.plot_pyqtgraph(show=False)
plt_ft.plot_pyqtgraph(show=True)

# plt_ic.plot_py(show=False, use_tex=False)
# plt_y21.plot_py(show=False, use_tex=False)
# plt_ft.plot_py(show=True, use_tex=False)

# plt_ic.save_tikz(path_data, standalone=True, build=True, clean=True, width="3in")
# plt_y21.save_tikz(path_data, standalone=True, build=True, clean=True, width="3in")
# plt_ft.save_tikz(path_data, standalone=True, build=True, clean=True, width="3in")



from pathlib import Path
from DMT.hl2 import McHicum, VA_FILES
from DMT.ADS import DutAds
from DMT.ngspice import DutNgspice
from DMT.core import Sweep, specifiers, SimCon, DutType, sub_specifiers, Plot

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

if __name__ == "__main__":
    mc = McHicum(load_model_from_path=folder_path / "model-card-examples.lib")
    mc.va_file = VA_FILES["L2V2.4.0_internal"]
    mc.set_values(
        {
            # 'flsh' :1,
            # 'alqav':0,
            # 'alkav':0,
            # # 'rth':1,
            # 'alrth':0,
            # 'alfav':0,
        }
    )
    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file
    dut_ngs = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        simulator_options={"GMIN": 1e-18, "ABSTOL": 1e-15},
    )

    plt_gum = Plot(
        r"$I_C(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "C" + "E",
        y_specifier=specifiers.CURRENT,
        y_log=True,
        style="mix",
        legend_location="upper right outer",
    )

    temps = [300, 350, 400, 450]
    for temp in temps:
        # create a sweep with ONE operating point
        sweepdef = [
            {
                "var_name": specifiers.VOLTAGE + "B",
                "sweep_order": 3,
                "sweep_type": "LIN",
                "value_def": [0.2, 1.2, 101],
            },
            {
                "var_name": specifiers.VOLTAGE + "C",
                "sweep_order": 2,
                "sweep_type": "CON",
                "value_def": [1],
            },
            {
                "var_name": specifiers.VOLTAGE + "E",
                "sweep_order": 1,
                "sweep_type": "CON",
                "value_def": [0],
            },
        ]
        outputdef = ["I_C", "I_B"]
        othervar = {"TEMP": temp}
        sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

        sim_con = SimCon(n_core=1, t_max=30)
        sim_con.append_simulation(dut=dut_ads, sweep=sweep)
        sim_con.append_simulation(dut=dut_ngs, sweep=sweep)
        sim_con.run_and_read(force=False)

        for key_ads, key_ngs in zip(dut_ads.data.keys(), dut_ngs.data.keys()):
            temp_df = dut_ads.get_key_temperature(key_ads)
            if not temp_df == temp:
                continue

            df_ads = dut_ads.data[key_ads]
            df_ads.ensure_specifier_column(specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED)
            df_ngs = dut_ngs.data[key_ngs]
            df_ngs.ensure_specifier_column(specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED)

            ic = df_ads[specifiers.CURRENT + "C"]
            ib = df_ads[specifiers.CURRENT + "B"]
            vbe = df_ads[specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED]
            plt_gum.add_data_set(vbe, ic, label="ads ic T={:f}".format(temp))
            plt_gum.add_data_set(vbe, ib, label="ads ib T={:f}".format(temp))

            ic = df_ngs[specifiers.CURRENT + "C"]
            ib = df_ngs[specifiers.CURRENT + "B"]
            vbe = df_ngs[specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED]
            # p   = df_ngs['Psh']
            t = df_ngs["DTSH"]
            plt_gum.add_data_set(vbe, ic, label="ngs ic T={:f}".format(temp))
            plt_gum.add_data_set(vbe, ib, label="ngs ib T={:f}".format(temp))

    plt_gum.plot_py(show=True)



""" test DutHdev
"""
import logging

from DMT.core import DutType, SimCon, constants
from DMT.Hdev import DutHdev
from hdevpy.tools import (
    get_sweep,
    get_hdev_dfs,
    get_default_plot,
)
from hdevpy.devices import n3_hbt
import numpy as np
from pathlib import Path

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=Path(__file__).resolve().parent.parent.parent / "logs" / "test_hdev.log",
    filemode="w",
)


def test_dut_hdev():
    # define a hdev input structure, look into the "n3_hbt" function for details
    # hint: have a look into the n3_hbt function to understand how a 1D HBT is defined in detail! Also check out the Hdev manual https://gitlab.com/metroid120/hdev_simulator
    inp = n3_hbt(
        "1.20", reco=True, tn=True
    )  # tn=True => Energy Transport Simulation, tn=False => Drift-Diffusion simulation

    # Create a DUT using DMT.core.DutHdev, which is used for simulation control
    dut = DutHdev(
        None,
        DutType.npn,
        inp,
        reference_node="E",
    )

    # define an output sweep using the hdevpy function get_sweep
    sweep = get_sweep(vce=[0, 9, 201], vbe=[0.65], ac=False)

    # start simulations
    simcon = SimCon(n_core=4, t_max=460)
    simcon.append_simulation(dut, sweep)
    simcon.run_and_read(force=False)

    # retrieve simulation data for sweep "sweep" => one pandas DataFrame with IV data, one with internal quantities "inqu"
    df_iv, df_inqu = get_hdev_dfs(dut, sweep, index=10)

    # get various data for tests
    ic = df_iv["I_C"].to_numpy()
    ie = df_iv["I_E"].to_numpy()
    ib = df_iv["I_B"].to_numpy()
    vb = df_iv["V_B"].to_numpy()

    n = df_inqu["N"].to_numpy()
    p = df_inqu["P"].to_numpy()
    ec = df_inqu["EC"].to_numpy()
    ev = df_inqu["EV"].to_numpy()
    efn = -df_inqu["PHI|N"].to_numpy()
    efp = -df_inqu["PHI|P"].to_numpy()
    nc = df_inqu["n_eff"].to_numpy()
    nv = df_inqu["p_eff"].to_numpy()
    vt = constants.P_K * df_inqu["TEMP"].to_numpy() / constants.P_Q

    n_dmt = nc * np.exp((efn - ec) / vt)
    p_dmt = nv * np.exp(-(efp - ev) / vt)

    assert np.isclose(vb, 0.65).all()  # output voltage makes sense
    assert np.isclose(ic[50:], -ie[50:] - ib[50:], rtol=1e-2).all()  # global continuity equation
    assert np.isclose(n, n_dmt, rtol=1e-3).all()  # n carrier densities make sense
    assert np.isclose(p, p_dmt, rtol=1e-3).all()  # p carrier densities make sense

    return dut, sweep


if __name__ == "__main__":
    dut, sweep = test_dut_hdev()



from pathlib import Path
from DMT.hl2 import McHicum, VA_FILES
from DMT.ADS import DutAds
from DMT.ngspice import DutNgspice
from DMT.core import Sweep, specifiers, SimCon, DutType, sub_specifiers, Plot

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

if __name__ == "__main__":
    mc = McHicum(load_model_from_path=folder_path / "dietmar_full_sh.lib")
    mc.va_file = VA_FILES["L2V2.4.0_internal"]
    mc.set_values(
        {
            "flsh": 1,
            # 'alqav':0,
            # 'alkav':0,
            # # 'rth':1,
            # 'alrth':0,
            # 'alfav':0,
        }
    )
    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file
    dut_ngs = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        simulator_options={"GMIN": 1e-18, "ABSTOL": 1e-15},
    )

    # create a sweep with ONE operating point
    sweepdef = [
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [0.9],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0, 3, 601],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    sim_con = SimCon(n_core=1, t_max=30)
    sim_con.append_simulation(dut=dut_ads, sweep=sweep)
    sim_con.append_simulation(dut=dut_ngs, sweep=sweep)
    sim_con.run_and_read(force=True)

    plt_out = Plot(
        r"$I_C(V_{\mathrm{CE}})$",
        x_specifier=specifiers.VOLTAGE + "C" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=False,
        style="mix",
        legend_location="upper right outer",
    )
    plt_p = Plot(
        r"$P(V_{\mathrm{CE}})$",
        x_specifier=specifiers.VOLTAGE + "C" + "E",
        y_specifier=specifiers.POWER,
        y_log=False,
        style="mix",
        legend_location="upper right outer",
    )
    plt_dT = Plot(
        r"$dT(V_{\mathrm{CE}})$",
        x_specifier=specifiers.VOLTAGE + "C" + "E",
        y_specifier=specifiers.TEMPERATURE,
        y_log=False,
        style="mix",
        legend_location="upper right outer",
    )

    for key_ads, key_ngs in zip(dut_ads.data.keys(), dut_ngs.data.keys()):
        df_ads = dut_ads.data[key_ads]
        df_ads.ensure_specifier_column(specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED)
        df_ngs = dut_ngs.data[key_ngs]
        df_ngs.ensure_specifier_column(specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED)

        ic = df_ads[specifiers.CURRENT + "C"]
        ib = df_ads[specifiers.CURRENT + "B"]
        vce = df_ads[specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED]
        p = df_ads["Psh"]
        t = df_ads["DTSH"]
        plt_out.add_data_set(vce, ic, label="ads ic")
        plt_out.add_data_set(vce, ib, label="ads ib")
        plt_p.add_data_set(vce, p, label="ads p")
        plt_dT.add_data_set(vce, t, label="ads dt")

        ic = df_ngs[specifiers.CURRENT + "C"]
        ib = df_ngs[specifiers.CURRENT + "B"]
        vce = df_ngs[specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED]
        p = df_ngs[specifiers.POWER]
        # p   = df_ngs['Psh']
        t = df_ngs["DTSH"]
        plt_out.add_data_set(vce, ic, label="ngs ic")
        plt_out.add_data_set(vce, ib, label="ngs ib")
        plt_p.add_data_set(vce, p, label="ngs p")
        plt_dT.add_data_set(vce, t, label="ngs dt")

    plt_dT.plot_py(show=False)
    plt_p.plot_py(show=False)
    plt_out.plot_py(show=True)



""" Base class to handle Verilog-AMS modelcard parameters.

Each parameter has a type, unit, boundaries and invalid values (excludes),
this is taken care of here.
Usually the user has a group of parameters stored in a Collection.
The collection exposes methods to manage the group safely. In generall here many copies are used,
in the exposes methods always deepcopies are returned and set to the collection. This reduces crazy errors, but also need to be handled with care.

Additionally parameters can be compared, they are considered equal, if they have the
same name and value. Also collections can be compared, they are equal, if they have
the same parameters and all parameters are equal.

Finally the classes here also add some pretty printing and loading and saving using pickle.
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from __future__ import annotations
import logging
import copy
import json
import warnings
from pint import Unit

try:
    from semver.version import Version as VersionInfo
except ImportError:
    from semver import VersionInfo


from pathlib import Path

import _pickle as cpickle  # type: ignore
import numpy as np
from typing import Dict, OrderedDict, Type, Union, List, Optional
from pint.formatting import siunitx_format_unit
from pint.errors import UndefinedUnitError

from DMT.core import unit_registry
from DMT.exceptions import (
    ValueExcludedError,
    ValueTooLargeError,
    ValueTooSmallError,
    BoundsError,
    ParaExistsError,
)

try:
    # try to import only, so it is a soft dependency...
    from pylatex import LongTable, MultiColumn, Section, NoEscape
    from DMT.external.pylatex import Tex  # pylint: disable=ungrouped-imports
except ImportError:
    pass

SEMVER_MCPARAMETER_CURRENT = VersionInfo(major=1, minor=0)
SEMVER_MCPARAMETER_Collection_CURRENT = VersionInfo(major=1, minor=0)


class McParameter(object):
    """Objects of this class represent a model card parameter. If you want to store many of them, see McParameterCollection class.

    Attributes
    ----------
    _value   :  Union[float, int, None]
        The value of this parameter.
    name     :  str
        The name of the parameter.
    inc_min  :  bool
        If True, value==min is allowed.
    inc_max  :  bool
        If True, value==max is allowed.
    min      :  Union[float, int]
        The minimum boundary of this parameter.
    max      :  Union[float, int]
        The maximum boundary of this parameter.
    exclude  :  Optional[List[Union[float, int]]]
        Optional list of values that can be excluded as a valid value for value. E.g. if min=-1, max=1, sometimes you might want to exclude 0.
    val_type : Type[Union[int, float]]
        The type of the value.
    description : str
        Description of the parameter

    Parameters
    ----------
    name : str
        Name of the parameter.
    value : Union[float, int, None]
        Value for the parameter. Can also be a other Parameter, then all attributes are copied.
    unit : pint.unit
        Unit of the python Pint package.
    minval : Union[float, int, None]
        Minimum boundary value of the parameter.
    maxval : Union[float, int, None]
        Maximum boundary value of the parameter.
    group  : str
        Display is sorted by groups.
    value_type : Type[Union[int, float]]
        The type of the value.
    inc_min  :  bool
        If True, value==min is allowed.
    inc_max  :  bool
        If True, value==max is allowed.
    exclude  :  Optional[List[Union[float, int]]]
        List of values that are excluded as a valid value for value. E.g. if min=-1, max=1, sometimes you might want to exclude 0.
    description : str
        Description of the parameter

    Methods
    -------
    check_bounds(value)
        Check wheather or not value is within the bounds of this parameter.
    _set_forced( value)
        Force setting the value. ATTENTION: When used, the boundaries may be set to inf!
    dict_json()
        Returns a compact formatted json dump of this parameter
    load_json(cls, name, value, __McParameter__, min, max, type, inc_min, inc_max, exclude, group, unit, description)
        Creates a McParameter from a dictionary obtained by a json.load.

    """

    def __init__(
        self,
        name: str,
        value: Optional[Union[float, int]] = None,
        minval: Optional[Union[float, int]] = None,
        maxval: Optional[Union[float, int]] = None,
        value_type: Type = float,
        inc_min: bool = True,
        inc_max: bool = True,
        exclude: Union[List[Union[float, int]], float, int, None] = None,
        group: str = "",
        unit: Unit = unit_registry.dimensionless,
        description: str = "",
    ):

        if not isinstance(name, str):
            raise IOError("DMT -> McParameter: Parameter name not a string.")
        self.name = name.lower()  # always lower case...
        self.inc_min = inc_min
        self.inc_max = inc_max
        if value_type == int:
            self._val_type = int
        elif value_type == float:
            self._val_type = float
        else:
            raise NotImplementedError(
                "The type "
                + str(value_type)
                + " of parameter value is not known! Allowed: int, float."
            )

        if (minval is None) or (minval == -np.inf):
            self._min = -np.inf
        else:
            self._min = self._val_type(minval)
        if (maxval is None) or (maxval == np.inf):
            self._max = np.inf
        else:
            self._max = self._val_type(maxval)

        self._exclude: List[Union[float, int]] = []
        self.exclude = exclude

        if value is None:
            self._value = None
        else:
            self._value = self._val_type(value)
        self.unit = unit
        self.group = group
        self.description = description

    def __repr__(self):
        """Set a better repr than McParameter Object at 0xTTTT

        Returns
        -------
        str
            This string could be passed to eval() to get a object with the same values as this one.
        """
        if self._val_type == int:
            str_type = "int"
        elif self._val_type == float:
            str_type = "float"
        else:
            str_type = str(self._val_type)  # make it reprable always...

        str_exclude = "[" + ";".join(f"{excluded:.5g}" for excluded in self.exclude) + "]"

        return (
            "McParameter("
            + self.name
            + f", value={self.value:g}"
            + f", minval={self._min:g}"
            + f", maxval={self._max:g}"
            + ", value_type="
            + str_type
            + ", inc_min="
            + str(self.inc_min)
            + ", inc_max="
            + str(self.inc_max)
            + ", exclude="
            + str_exclude
            + ', group="'
            + self.group
            + '", unit=unit_registry("'
            + str(self.unit)
            + '").units, description="'
            + self.description
            + '")'
        )

    def dict_json(
        self,
    ) -> dict[str, Union[float, int, str, bool, None, List[Union[float, int]]]]:
        """Returns a compact formatted json dump of this parameter"""

        if self._val_type == int:
            str_type = "int"
        elif self._val_type == float:
            str_type = "float"
        else:
            str_type = str(self._val_type)  # make it saveable always...

        try:
            desc = self.description
        except AttributeError:
            desc = ""

        # str_exclude = "[" + ";".join(f"{excluded:.5g}" for excluded in self.exclude) + "]"

        return {
            "name": self.name,
            "value": self._value,
            "min": self._min,
            "max": self._max,
            "inc_min": self.inc_min,
            "inc_max": self.inc_max,
            "exclude": self.exclude,
            "type": str_type,
            "unit": str(self.unit),
            "group": self.group,
            "description": desc,
            "__McParameter__": str(
                SEMVER_MCPARAMETER_CURRENT
            ),  # make versions, so we can introduce compatibility here!
        }

    @classmethod
    def load_json(
        cls,
        name: str,
        value: Union[float, int],
        __McParameter__: Union[float, str],
        min: Optional[Union[float, int]] = None,
        max: Optional[Union[float, int]] = None,
        type: str = "",
        inc_min: bool = True,
        inc_max: bool = True,
        exclude: Optional[List[Union[float, int]]] = None,
        group: str = "",
        unit: Union[str, Unit] = "",
        description: str = "",
    ):
        """Creates a McParameter from a dictionary obtained by a json.load."""
        if isinstance(__McParameter__, float):
            __McParameter__ = (
                f"{__McParameter__:1.1f}.0"  # if it is a number only MAJOR.MINOR is used
            )

        if VersionInfo.parse(__McParameter__) == SEMVER_MCPARAMETER_CURRENT:
            try:
                value_type = {"int": int, "float": float}[type]
            except KeyError:
                value_type = type

            if isinstance(unit, str):
                try:
                    unit = unit_registry(unit).units
                except UndefinedUnitError:
                    unit = unit_registry.dimensionless

            return McParameter(
                name,
                value=value,
                unit=unit,
                minval=min,
                maxval=max,
                group=group,
                value_type=value_type,
                inc_min=inc_min,
                inc_max=inc_max,
                exclude=exclude,
                description=description,
            )
        else:
            raise ValueError(
                f"DMT->McParameter: This dict has an unknown McParameter json version: {__McParameter__:f}"
            )

    @property
    def val_type(self):
        """Return the type of the value."""
        return self._val_type

    @val_type.setter
    def val_type(self, new_type):
        """Set the type of this parameter."""
        if self._value is None:
            self._val_type = new_type  # nothing to do here
        elif new_type == int:
            if self._value == int(self._value):  # test if roundable...
                self._value = int(self._value)
                self._val_type = int
            else:
                raise IOError(
                    "The parameter value was a floating number and it was tried to set the parameter type to integer. The parameter name is: "
                    + self.name
                )
        elif new_type == float:
            self._value = float(self._value)  # can be set always
            self._val_type = new_type
        else:
            raise IOError(
                "This type can not be set for McParameter: "
                + str(new_type)
                + ". The parameter name is: "
                + self.name
            )

    @property
    def min(self) -> Union[float, int]:
        """Return The minimum boundary as an array of length one."""
        return self._min

    @min.setter
    def min(self, min_new: Union[float, int]):
        """Set the minimum boundary and throw errors if min>value or min>max, testing inc_min before doing so."""
        if min_new > self.max:
            raise BoundsError(
                "DMT -> McParameter: The new minimum is above the maximum of the parameter"
            )

        if self._value is None:
            pass
        elif self.inc_min:
            if min_new > self._value:
                raise BoundsError(
                    "DMT -> McParameter: Parameter min value of "
                    + self.name
                    + " can not be set to "
                    + str(min_new)
                    + " since value is currently "
                    + str(self.value)
                    + " ."
                )
        else:
            if min_new >= self._value:
                raise BoundsError(
                    "DMT -> McParameter: Parameter min value of "
                    + self.name
                    + " can not be set to "
                    + str(min_new)
                    + " since value is currently "
                    + str(self.value)
                    + " ."
                )

        self._min = self.val_type(min_new)

    @property
    def max(self) -> Union[float, int]:
        """Return The minimum boundary as an array of length one."""
        return self._max

    @max.setter
    def max(self, max_new: Union[float, int]):
        """Set the max boundary and throw errors if max<value or max<max, testing inc_max before doing so."""
        if max_new < self.min:
            raise BoundsError(
                "DMT -> McParameter: The new minimum is above the maximum of the parameter"
            )

        if self._value is None:
            pass
        elif self.inc_max:
            if max_new < self._value:
                raise BoundsError(
                    "DMT -> McParameter: Parameter max value of "
                    + self.name
                    + " can not be set to "
                    + str(max_new)
                    + " since value is currently "
                    + str(self.value)
                    + " ."
                )
        else:
            if max_new <= self._value:
                raise BoundsError(
                    "DMT -> McParameter: Parameter max value of "
                    + self.name
                    + " can not be set to "
                    + str(max_new)
                    + " since value is currently "
                    + str(self.value)
                    + " ."
                )

        self._max = self.val_type(max_new)

    @property
    def value(self) -> Union[float, int, None]:
        """Returns the value."""
        if self._value is None:
            return None

        return self.val_type(self._value)

    @value.setter
    def value(self, value: Union[float, int, McParameter]):
        if isinstance(value, McParameter):
            self.name = value.name
            self._min = value._min  # pylint: disable=protected-access
            self._max = value._max  # pylint: disable=protected-access
            self.inc_max = value.inc_max
            self.inc_min = value.inc_min
            self.exclude = value.exclude
            self._val_type = value.val_type
            # self.value = value.value

            value = self.check_bounds(value.value)  # type: ignore
            self._value = value
        else:
            value = self.check_bounds(value)
            self._value = value

    @property
    def exclude(self):
        """Return the type of the value."""
        return self._exclude

    @exclude.setter
    def exclude(self, new_exclude: Union[List[Union[float, int]], float, int, None]):
        """Set the type of this parameter."""
        if new_exclude is None:
            self._exclude = []
        else:
            try:
                self._exclude = [self._val_type(val) for val in new_exclude]  # type: ignore
            except TypeError:
                self._exclude = [self._val_type(new_exclude)]  # type: ignore

    def _set_forced(self, value: Union[float, int]):
        """Force setting the value. ATTENTION: When used, the boundaries may be set to inf!"""
        try:
            # try without changing bounds
            self.value = value
        except (ValueTooLargeError, ValueTooSmallError, ValueExcludedError) as err:
            if isinstance(value, McParameter):
                raise IOError(
                    "McParameter _set_forced: The given McParameter is already inconsistent. Stop."
                ) from err
            else:
                # set to no bounds
                if self.val_type == float:
                    self._min = -np.inf
                    self._max = np.inf
                elif self.val_type == int:  # inf not possible in integer...
                    self._min = np.iinfo(int).min
                    self._max = np.iinfo(int).max
                else:
                    raise NotImplementedError("This type is not Implemented") from err
                self._value = self.val_type(value)

    def check_bounds(self, value: Union[float, int]):
        """Check wheather the value parameter is inside the boundaries defined by self.min and self.max.

        Parameters
        ----------
        value  :  int or float or convertable to float
            Value that shall be checked.

        Returns
        -------
        value  :  int or float
            Checked value
        """
        # type check, either int or float is allowed
        if self.val_type == int:
            if int(value) != value:
                raise TypeError(f"The parameter {self:s} is of type Integer!")

            value = int(value)
        elif not isinstance(
            value, (int, float)
        ):  # for floats also integer are allowed. This catches everything else like strings or lists etc.
            raise TypeError(f"The parameter {self:s} is of type Float!")

        # range check
        value_too_large = False
        value_too_small = False

        if self.inc_min and value < self.min:
            value_too_small = True
        elif not self.inc_min and value <= self.min:
            value_too_small = True

        if self.inc_max and value > self.max:
            value_too_large = True
        elif not self.inc_max and value >= self.max:
            value_too_large = True

        if value_too_large:
            raise ValueTooLargeError(
                f"Value of {self:s} above its maximum! Given: {value:e}! Maximum boundary : {self.max:e}!"
            )

        if value_too_small:
            raise ValueTooSmallError(
                f"Value of {self:s} below its minimum! Given: {value:e}! Minimum boundary : {self.min:e}!"
            )

        # exclude check
        if self.exclude is not None:
            if value in self.exclude:
                str_excluded = ";".join(f"{excluded:g}" for excluded in self.exclude)
                raise ValueExcludedError(
                    f"Value of {self:s} is excluded! Given: {value:e}! Excluded : [{str_excluded:s}]!"
                )

        return value

    def __format__(self, wanted_format: str) -> str:
        """Allows formating of McParameters using "{}".format(mc_parameter).

        If a number format (defg) is given, the value is formated, for strings (s) the name.
        Additionally the unit (u) in siunitx format is possible.
        """
        if (
            ("d" in wanted_format)
            or ("e" in wanted_format)
            or ("f" in wanted_format)
            # or ("g" in wanted_format)
        ):
            if self.value is None:
                return "-"  # dirty

            return f"{self.value:{wanted_format}}"

        if "g" in wanted_format:
            if self.value is None:
                return "-"  # dirty

            if self.val_type == float:
                return f"{self.value:{wanted_format}}"
            else:
                # 10.5g -> 10d
                wanted_format = wanted_format.replace("g", "d")

                return f"{self.value:{wanted_format}}"

        if "s" in wanted_format:
            return f"{self.name:{wanted_format}}"

        if "u" in wanted_format:
            if hasattr(self, "unit") and self.unit is not None and not self.unit.dimensionless:
                # return siunitx_format_unit(self.unit)
                try:
                    return siunitx_format_unit(self.unit)  # type: ignore
                except TypeError:
                    return siunitx_format_unit(
                        self.unit._units, unit_registry
                    )  # new version has other interface
            else:
                return "-"

        raise IOError(f"The format {wanted_format} is unknown for McParameters!")

    def __eq__(self, other: McParameter) -> bool:
        """Comparing parameters, equal if name and value is equal."""
        if isinstance(other, McParameter):
            return (self.name == other.name) and (self.value == other.value)

        return NotImplemented


class McParameterCollection(object):
    """
    This parameter collection has properties which as a single parameter. This way a group of parameters and a single parameter can be treated equally.

    Attributes
    ----------
    paras : list
        The parameters of this group

    Parameters
    ----------
    possible_groups : dict[str, str], optional
        Dictionary of possible groups in this collection, saved as Description: GroupName, by default None
    __McParameterCollection__ : Union[VersionInfo, str, float], optional
        Version of the given creation parameters, by default SEMVER_MCPARAMETER_Collection_CURRENT

    Raises
    ------
    NotImplementedError
        Raised when the given version is unknown and hence the
    """

    def __init__(
        self,
        possible_groups: Optional[Dict[str, str]] = None,
        __McParameterCollection__: Union[
            VersionInfo, str, float
        ] = SEMVER_MCPARAMETER_Collection_CURRENT,
        **kwargs,
    ):
        if not isinstance(__McParameterCollection__, VersionInfo):
            try:
                __McParameterCollection__ = VersionInfo.parse(__McParameterCollection__)  # type: ignore
            except TypeError:
                __McParameterCollection__ = VersionInfo.parse(
                    f"{__McParameterCollection__:1.1f}.0"
                )  # if it is a number only MAJOR.MINOR is used

        if __McParameterCollection__ != SEMVER_MCPARAMETER_Collection_CURRENT:
            raise NotImplementedError(
                "DMT->McParameterCollection: Unknown version of collection to create!"
            )

        self._paras: list[McParameter] = list()
        self._values: OrderedDict[str, Union[float, int]] = OrderedDict()

        if possible_groups is None:
            self.possible_groups = {}
        else:
            self.possible_groups = possible_groups

        for key in kwargs:
            warnings.warn(
                f"{type(self)} was created with the remaining kwarg {key}={kwargs[key]}",
                category=RuntimeWarning,
            )

    @property
    def paras(self):
        """Return the parameters with updated values."""
        self.update_paras()
        return self._paras

    @paras.setter
    def paras(self, paras_new):
        """Set the parameters."""
        self._paras = paras_new
        self.update_values()

    def update_paras(self):
        """Writes back the values into the parameters."""
        for parameter in self._paras:
            parameter.value = self._values[parameter.name]

    def update_values(self):
        """Writes the parameter values into the values dict."""
        self._values = OrderedDict()
        for parameter in self._paras:
            self._values[parameter.name] = parameter.value  # type: ignore

    def info_json(self, **_kwargs):
        """Returns a dict with serializeable content for the json file to create. Add the info about the concrete subclass to create here!"""
        return {
            "possible_groups": self.possible_groups,
            "__McParameterCollection__": str(SEMVER_MCPARAMETER_Collection_CURRENT),
        }  # make versions, so we can introduce compatibility here!

    def dump_json(self, file_path, **kwargs):
        """Writes itself and the parameters in the collection to a file.

        To manipulate what is written to the file, change :py:method::`DMT.core.mc_parameter.McParameterCollection.dumps_json()`
        """
        if not isinstance(file_path, Path):
            file_path = Path(file_path)

        content = []
        for para in sorted(self.paras, key=lambda x: (x.group, x.name)):
            content.append(para.dict_json())
        content.append(self.info_json(**kwargs))

        str_content = "["
        for subdict in content:
            str_content += "\n    " + json.dumps(subdict) + ","
        # remove trailing ,
        str_content = str_content[:-1]
        str_content += "]"

        file_path.write_text(str_content, encoding="utf8")

    @classmethod
    def load_json(
        cls,
        file_path: Union[str, Path],
        directory_va_file: Union[str, Path, None] = None,
        ignore_checksum: bool = False,
    ) -> McParameterCollection:
        """Loads the json file, creates the McParameterCollection (or inherited) and adds the McParameters.

        Parameters
        ----------
        file_path : Union[str, Path]
            Path to the json.
        directory_va_file : Union[str, Path, None], optional
            If a relative path to a va_file is set in the modelcard, pass the absolute path to the start folder here, by default None.
            This can be used to load old json modelcards from before saving the full code with the parameters.
        ignore_checksum : bool, optional
            When the code is saved compressed, a checksum is saved with it. If you want to ignore the checksum set this to true, by default False


        Returns
        -------
        McParameterCollection
            The loaded collection

        Raises
        ------
        IOError
            If the collection dictionary is not found in the json file.
        """
        if not isinstance(file_path, Path):
            file_path = Path(file_path)

        with file_path.open("r", encoding="utf8") as file_json:
            content = json.load(file_json)

        collection = None
        for dict_content in content:
            if (
                "__McParameterCollection__" in dict_content
                or "__McParameterComposition__" in dict_content
            ):
                collection = cls(
                    directory_va_file=directory_va_file,
                    ignore_checksum=ignore_checksum,
                    **dict_content,
                )
                break

        if collection is None:
            raise IOError(
                "DMT->McParameterCollection: Did not create the collection as no collection dictionary is found in the file.",
                "Try to create a collection object yourself and the load the parameter json files manually.",
            )

        for dict_content in content:
            if "__McParameter__" in dict_content:
                try:
                    collection.add(McParameter.load_json(**dict_content), update=False)
                except ParaExistsError:
                    collection.set(McParameter.load_json(**dict_content), update=False)

        collection.update_values()

        return collection

    def get(
        self, parameters: Union[str, McParameter, list[str], tuple[str], McParameterCollection]
    ) -> Union[McParameter, McParameterCollection]:
        """Returns a McParameterCollection with copies of all given parameter names.

        Parameters
        ----------
        parameters  : str, iterable(str) or McParameterCollection

        Returns
        -------
        mcard_collection : McParameterCollection

        Raises
        ------
        KeyError
            If the para was not in self.
        """
        if isinstance(parameters, (McParameterCollection, list, tuple)):
            mcard_collection = McParameterCollection()
            for para in parameters:
                mcard_collection.add(self.get(para), update=False)

            mcard_collection.update_values()
            return mcard_collection

        try:
            if isinstance(parameters, McParameter):
                my_para = next(para for para in self._paras if para.name == parameters.name.lower())
            elif isinstance(parameters, str):
                my_para = next(para for para in self._paras if para.name == parameters.lower())
            else:
                raise IOError("The parameter is neither of type McParameter or str.")

        except StopIteration as err:
            raise KeyError(
                f"The parameter {parameters:s} is not part of this parameter collection!"
            ) from err

        my_para.value = self._values[my_para.name]
        return copy.deepcopy(my_para)

    def __getitem__(self, para):
        """Allows paras['c10']"""
        return self.get(para)

    def set(self, parameters, update=True, force=False):
        """Set existing paramaters in self.

        Parameters
        ----------
        parameters : McParameter or McParameterCollection
            For each parameter, if it is found in self, it is removed and the given is added. If it is not found, a KeyError is raised.
        update : {True, False}, optional
            If set to False, the values dict is not updated.
        force :  {True, False}, optional
            If set to True, the parameter is added if it did not exist.

        Raises
        ------
        KeyError
            If the para was not in self.
        """
        if isinstance(parameters, McParameterCollection):
            for para in parameters:
                self.set(para, force=force, update=update)
            return

        try:
            index = self.name.index(parameters.name)
        except ValueError as err:
            if force:
                index = -1
            else:
                raise KeyError(
                    f"The parameter {parameters:s} is not part of this parameter collection!"
                ) from err

        if index >= 0:
            self.remove(parameters.name)

        self.add(parameters, index=index, update=False)

        if update:
            self._values[parameters.name] = parameters.val_type(parameters.value)

    def __setitem__(self, name, value):
        """Allows paras['c10']"""
        para = self[name]
        para.value = value
        self.set(para)

    def set_values(self, dict_parameters, force=False, policy_missing="error"):
        """Sets a dictionary of {'name':value} to the parameters in this collection

        Parameters
        ----------
        dict_parameters : {str: float64}
            For each parameter, if it is found in self, the given value is set.
        force : boolean, optional
            If True, values are force set.
        policy_missing : {"error", "ignore", "add"}, optional
            The policy for missing parameters to set, defaults to "error"

        Raises
        ------
        KeyError
            If the para was not in self.
        """
        for name, value in dict_parameters.items():
            name = name.lower()
            try:
                index = self.name.index(name)
            except ValueError as err:
                if policy_missing == "ignore":
                    # nothing to do here
                    continue
                elif policy_missing == "add":
                    para = McParameter(name=name, value=value)
                    self.add(para)
                    index = self.name.index(name)
                else:
                    raise KeyError(
                        f"The parameter {name:s} is not part of this parameter collection!"
                    ) from err

            try:
                self._paras[index].value = value
            except ValueError:
                if force:
                    para = self.get(name)
                    para._set_forced(value)  # type: ignore
                    self.set(para, update=False)
                else:
                    raise

            dict_parameters[name] = self._paras[index].val_type(value)

        self._values.update(dict_parameters)

    def get_values(self, parameters):
        """Returns a list of the values of parameters.

        Returns
        -------
        {name:value}
            A dict with the name of the parameter as key and value as value.

        Raises
        ------
        KeyError
            If the para was not in self.
        """
        values = {}
        for name in parameters:
            values[name] = self._values[name.lower()]

        return values

    def set_bounds(self, dict_parameters):
        """Sets a dictionary of {'name':(min, max )} to the parameters in this collection

        Parameters
        ----------
        dict_parameters : {str: (float64, float64)}
            For each parameter, if it is found in self, the given values are set as minimum and maximum.

        Raises
        ------
        KeyError
            If the para was not in self.
        """
        for name, values in dict_parameters.items():
            try:
                index = self.name.index(name.lower())
            except ValueError as err:
                raise KeyError(
                    f"The parameter {name:s} is not part of this parameter collection!"
                ) from err

            self.paras[index].min = values[0]
            self.paras[index].max = values[1]

    def to_kwargs(self):
        """Returns itself as a dictionary fitting to unpack into a function call.

        Returns
        -------
        dict
            {name: value}
        """
        # dict_a = {}
        # for para in self.paras:
        #     dict_a[para.name] = para.value

        return self._values

    def print_parameters(self, paras=None, line_break=""):
        """Just some pretty printing

        Parameters
        ----------
        param : list[str], optional
            List of parameter names to print, if not given, all children are returned!
        line_break : str, optional
            Is added after each parameter, can be used as line breaks

        Returns
        -------
        str
            String with all parameters.
        """
        temp_str = ""
        if paras is None:
            # if None iterate through all
            paras = self.paras
        else:
            paras = self.get(paras)

        try:
            for para in sorted(paras, key=lambda x: (x.group, x.name)):  # type: ignore
                temp_str += f"  {para:<12s} = {para:10.5e} {line_break}"
        except TypeError:  # no groups available
            for para in sorted(paras, key=lambda x: x.name):  # type: ignore
                temp_str += f"  {para:<12s} = {para:10.5e} {line_break}"

        return temp_str

    def print_to_file(self, path_to_file, line_break="", create_dir=False):
        """Prints the parameters into a file. Uses :meth:`print_parameters` to obtain the string to print.

        Parameters
        ----------
        path_to_file : str
            Path to the file to write. '.txt' is added automatically.
        line_break : str, optionally
            Is added after each parameter, can be used as line breaks
        create_dir : {False, True}, optionally
            If true, the respective directory is created first.
        """
        if not isinstance(path_to_file, Path):
            path_to_file = Path(path_to_file)

        if path_to_file.suffix != ".txt":
            path_to_file = path_to_file.with_suffix(".txt")

        if create_dir:
            path_to_file.parent.mkdir(parents=True, exist_ok=True)

        path_to_file.write_text(
            self.print_parameters(line_break=line_break) + "\n", encoding="utf8"
        )

    @classmethod
    def load(cls, path):
        """Load an object from a pickle file.

        Parameters
        ----------
        path : str
            Path to the file to load.

        Returns
        -------
        McParameterCollection
        """
        if not isinstance(path, Path):
            path = Path(path)

        with path.open("rb") as my_db:
            collection = cpickle.load(my_db)

        return collection

    @property
    def name(self):
        """Returns all names of the parameters in the collection"""
        # names = np.chararray(len(self.paras), itemsize=20)
        # for i in range(len(self.paras)):
        #     names[i] = self.paras[i].name
        return [para.name for para in self._paras]

    @property
    def group(self):
        """Returns all groups of the parameters in the collection as a set"""
        groups = []
        for para in self._paras:
            try:
                groups.append(para.group)
            except AttributeError:
                groups.append("")

        return set(groups)

    @property
    def unit(self):
        """Returns all units in the collection"""
        units = []
        for para in self._paras:
            try:
                units.append(para.unit)
            except AttributeError:
                units.append(unit_registry.dimensionless)

        return units

    @property
    def value(self):
        """Returns all parameter values as a np.ndarray."""
        vals = np.empty(len(self._paras))

        for i_para, para in enumerate(self._paras):
            vals[i_para] = para.value

        return vals

    @value.setter
    def value(self, value):
        """Sets all values for all Parameters. Value is a list, the children are set in the given order."""
        if len(value) != len(self._paras):
            raise IOError(
                "The amount of values to set must be the same as the amount of parameters in the collection!"
            )

        for para, val in zip(self._paras, value):
            para.value = val

        self.update_values()

    def remove(self, parameters):
        r"""Removes the given parameter names from the parameter collection.

        Parameters
        ----------
        parameters : str, iterable(str), McParameter or McParameterCollection

        """
        if isinstance(parameters, (list, tuple)):
            for para in parameters:
                self.remove(para)
            return
        elif isinstance(parameters, McParameterCollection):
            for para in parameters:
                self.remove(para.name)
            return

        if isinstance(parameters, McParameter):
            parameters = parameters.name  # extract the name

        try:
            i_para = next(i for i, my_para in enumerate(self._paras) if my_para.name == parameters)
        except StopIteration as err:
            raise KeyError(
                f"The parameter {parameters:s} is not part of this parameter collection and can not be removed!"
            ) from err

        del self._paras[i_para]
        del self._values[parameters]

    @property
    def min(self):
        """All minimal values of this group"""
        vals = np.empty(len(self))
        for i_para, para in enumerate(self._paras):
            vals[i_para] = para.min

        return vals

    @min.setter
    def min(self, min_new):
        """Sets all minimal values, sets each minimum specifically"""
        if len(min_new) != len(self):
            raise IOError(
                "The amount of minimum boundaries to set must be the same as the amount of parameters in the collection!"
            )

        for para, min_a in zip(self._paras, min_new):
            para.min = min_a

    @property
    def max(self):
        """All maximal values of this group"""
        vals = np.empty(len(self))
        for i_para, para in enumerate(self._paras):
            vals[i_para] = para.max

        return vals

    @max.setter
    def max(self, max_new):
        """Sets all maximal values, sets each maximum specifically"""
        if len(max_new) != len(self):
            raise IOError(
                "The amount of minimum boundaries to set must be the same as the amount of parameters in the collection!"
            )

        for para, max_a in zip(self._paras, max_new):
            para.max = max_a

    def print_tex(self, pretext="The final modelcard is summarized in the table below:"):
        """Prints a modelcard as a tex table using PyLaTeX"""
        # try to clean first
        try:
            clean_mcard = self.get_clean_modelcard()  # type: ignore
        except (
            AttributeError,
            KeyError,
            ValueError,
        ):  # was a broad except (add more types if needed)
            clean_mcard = self

        doc = Tex()
        # Generate data table
        with doc.create(Section("Modelcard")):
            doc.append(pretext)
            with doc.create(
                LongTable(NoEscape(r"l S"), width=2, booktabs=True)
            ) as data_table:  # pylatex does not count s S columns from siunitx
                data_table.add_hline()
                data_table.add_row(["parameter", NoEscape("{value}")])
                data_table.add_hline()
                data_table.end_table_header()
                data_table.add_hline()
                data_table.add_row(
                    (MultiColumn(2, align="r", data="continued on next Page"),), strict=False
                )
                data_table.add_hline()
                data_table.end_table_footer()
                data_table.add_hline()
                data_table.add_row((MultiColumn(2, align="r", data="Finish"),), strict=False)
                data_table.add_hline()
                data_table.end_table_last_footer()

                group = None
                for para in sorted(clean_mcard, key=lambda x: (x.group, x.name)):
                    if group != para.group:
                        if group is not None:
                            data_table.add_hline()  # horizontal line after each group and then the next group name

                        group = para.group
                        try:
                            group_desc = next(
                                description
                                for description, group_a in self.possible_groups.items()
                                if group_a == group
                            )
                            data_table.add_row(
                                (MultiColumn(2, align="l", data=group_desc),), strict=False
                            )
                        except StopIteration:
                            pass

                    if para.unit.dimensionless:
                        data_table.add_row(
                            [
                                NoEscape(f"{para:<12s}"),
                                NoEscape(f"{para:g}"),
                            ],
                            strict=False,
                        )
                    else:
                        data_table.add_row(
                            [
                                NoEscape(f"{para:<12s}/\si{{{para:u}}}"),
                                NoEscape(f"{para:g}"),
                            ],
                            strict=False,
                        )

        return doc

    def __iter__(self):
        # return iter(self.paras)
        return iter(copy.deepcopy(self.paras))

    def sort_paras(self):
        """Sorts the parameters according to the groups."""
        self._paras.sort(key=lambda x: (x.group, x.name))

    def iter_alphabetical(self):
        """Returns an iterator on parameters sorted alphabetically by name"""
        return iter(sorted(copy.deepcopy(self.paras), key=lambda para: para.name))

    def __len__(self):
        return len(self._paras)

    def __contains__(self, other):
        return other.name in self.name

    def add(self, paras, index=None, update=True):
        """Add a parameter to self. This is only allowed, if the parameter name is not already known to the collection."""
        if isinstance(paras, (McParameterCollection)):
            if index is None:
                for para in paras._paras:  # deepcopy is in the McParameter add
                    self.add(para, update=update)
            else:
                for para in paras._paras[
                    ::-1
                ]:  # reverse order if index is given -> insert turns the order around again
                    self.add(para, index=index, update=update)
            return

        if isinstance(paras, McParameter):
            if paras.name in self.name:
                raise ParaExistsError(
                    f"Tried to set parameter {paras:s}, which was already defined."
                )
            else:
                if index is None:
                    self._paras.append(copy.deepcopy(paras))
                else:
                    self._paras.insert(index, copy.deepcopy(paras))

                if update:
                    self._values[paras.name] = paras.val_type(paras.value)  # type: ignore
        else:
            raise TypeError(
                "McParameterCollection accepts only McParameter or McParameterCollection!"
            )

    def __add__(self, other):
        """Allows appending of two collections by mc1 + mc2"""
        if isinstance(other, (McParameter, McParameterCollection)):
            mc_return = copy.deepcopy(self)
            mc_return.add(other)

            return mc_return
        else:
            return NotImplemented

    def __radd__(self, other):
        """Called when parameter + collection is used. Here we need to take care of the index!"""
        if isinstance(other, (McParameter, McParameterCollection)):
            mc_return = copy.deepcopy(self)
            mc_return.add(other, index=0)  # insert at start

            return mc_return
        else:
            return NotImplemented

    def eq_paras(self, other):
        """Compares the parameters in two McParameterCollections or subclasses"""
        str_diff_vars = ""
        for para in self.paras:
            try:
                if para.value != other.get(para.name).value:
                    str_diff_vars += f"{para:<12s}: {para:10.4e} || {other.get(para):10.4e}\n"
            except KeyError:
                str_diff_vars += f"The second modelcard does not have a {para:s} parameter!\n"

        # find parameters in other which are not in self!
        for para in other:
            if para.name not in self.name:
                str_diff_vars += f"The first modelcard does not have a {para:s} parameter!\n"

        if str_diff_vars:
            logging.info(str_diff_vars)
            return False

        return True

    def __eq__(self, other):
        """Allows comparing 2 model cards using mc1 == mc2"""
        if isinstance(other, McParameterCollection):
            # can only compare to other collections
            return self.eq_paras(other)

        return NotImplemented


class McParameterComposition(McParameterCollection):
    """Deprecated name for the mc parameter collection"""

    def __init__(
        self,
        __McParameterComposition__: Union[
            VersionInfo, str, float
        ] = "1.0.0",  # as until now there is only v1.0.0 around we have no issues here.
        **kwargs,
    ):
        warnings.warn(
            "McParameterComposition is deprecated. It was renamed to McParameterCollection to avoid confusion with the composition design pattern.\nMcParameterCompostion will be deleted in the next major release.",
            category=DeprecationWarning,
        )

        super().__init__(**kwargs)



""" Testing to obtain the parameters of a function
"""
from pathlib import Path

from DMT.core import MCard
from DMT.external import get_param_list


def calculate_be_capacitance(model, vbe, t_dev, *, cbepar=None, flsh=None, **model_parameters):
    flsh = int(flsh)  # be save
    cjep = model.functions["Cjep"].eval(
        temperature=t_dev, voltages={"br_bpei": vbe}, flsh=flsh, **model_parameters
    )
    cjei = model.functions["Cjei"].eval(
        temperature=t_dev, voltages={"br_biei": vbe}, flsh=flsh, **model_parameters
    )
    return cjei + cjep + cbepar


def test_vae_module():
    mc_hicum_l2 = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
        va_file=Path(__file__).resolve().parent / "hicumL2V2p4p0.va",
    )
    mc_hicum_l2.update_from_vae(remove_old_parameters=True)
    module = mc_hicum_l2.get_verilogae_model()

    params = get_param_list(module.functions["Cjep"], info=None)

    assert set(params) == {
        "cjep0",
        "vdep",
        "vge",
        "zep",
        "ajep",
        "vgb",
        "f1vg",
        "flsh",
        "tnom",
        "type",
        "rth",
        "dt",
        "flcomp",
    }


def test_custom_fct():
    mc_hicum_l2 = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
        va_file=Path(__file__).resolve().parent / "hicumL2V2p4p0.va",
    )
    mc_hicum_l2.update_from_vae(remove_old_parameters=True)
    module = mc_hicum_l2.get_verilogae_model()

    model_function_info = {
        "independent_vars": ("vbe", "t_dev"),
        "depends": (module.functions["Cjep"], module.functions["Cjei"]),
    }
    params = get_param_list(calculate_be_capacitance, info=model_function_info)

    assert set(params) == {
        "ajep",
        "vge",
        "type",
        "cjep0",
        "cjei0",
        "ajei",
        "rth",
        "flsh",
        "vdei",
        "tnom",
        "vgb",
        "vdep",
        "zei",
        "f1vg",
        "cbepar",
        "zep",
        "dt",
        "flcomp",
    }


if __name__ == "__main__":
    test_vae_module()
    test_custom_fct()



import os
from pathlib import Path
from DMT.core import Sweep
from DMT.core import DutDummy
from DMT.core import SimCon
from DMT.core import DutType
from DMT.core import specifiers
from DMT.core import MCard


folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent


def test_run_sim():
    """Creates a sweep and a dummy dut to run using the simulation controller."""
    sim_con = SimCon(n_core=4, t_max=5)

    sweepdef = [
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 1,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [1],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 3,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    othervar = {"TEMP": 300, "w": 10, "l": 0.25}
    sweep = Sweep("test_sweep", sweepdef=sweepdef, othervar=othervar)

    mcard = MCard(
        ["A"],
        "Q_HIC",
        "hicuml2va",
        va_file=folder_path / "test_va_code" / "hicuml2" / "hicumL2V2p4p0_release.va",
    )
    mcard.load_model_parameters(
        folder_path / "test_modelcards" / "IHP_ECE704_03_para_D21.mat",
    )

    dut = DutDummy(
        test_path, "dummy", DutType.npn, mcard, nodes=None, reference_node="E", force=True
    )

    for _i in range(10):
        sim_con.append_simulation(dut, sweep)

    sim_con.run_and_read(force=True)

    (test_path / "iv.p").unlink()


if __name__ == "__main__":
    test_run_sim()
    dummy = 1



import pytest
from DMT.core import SimCon


def test_create_singleton():
    """Creates a sweep and a dummy dut to run using the simulation controller."""
    sim_con = SimCon(n_core=4, t_max=5)

    sim_con.run_and_read(force=True)

    assert sim_con.n_core == 4


def test_create_new_singleton():
    sim_con = SimCon(n_core=4, t_max=5)

    assert sim_con.n_core == 4

    id_old = id(sim_con)

    sim_con = SimCon(
        n_core=5
    )  # instanziation of "new" SimCon with different n_core leads to same object with changed n_core

    assert sim_con.t_max == 5
    assert sim_con.n_core == 5

    assert id(sim_con) == id_old


def test_reinstanziation_errors():
    _sim_con = SimCon(n_core=4, t_max=5)

    with pytest.raises(AttributeError):
        _sim_con = SimCon(5)

    with pytest.raises(AttributeError):
        _sim_con = SimCon(alternative_attr=5)


if __name__ == "__main__":
    test_create_singleton()
    test_create_new_singleton()
    test_reinstanziation_errors()
    dummy = 1
